"""Objects representing API generators to MediaWiki site."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import heapq
import itertools
import typing
from contextlib import suppress
from itertools import zip_longest
from typing import Any

import pywikibot
from pywikibot.backports import Callable, Generator, Iterable, batched
from pywikibot.data import api
from pywikibot.exceptions import (
    APIError,
    Error,
    InconsistentTitleError,
    InvalidTitleError,
    NoPageError,
    UserRightsError,
)
from pywikibot.site._decorators import need_right
from pywikibot.site._namespace import NamespaceArgType
from pywikibot.tools import (
    deprecate_arg,
    deprecated,
    deprecated_signature,
    is_ip_address,
)
from pywikibot.tools.itertools import filter_unique


if typing.TYPE_CHECKING:
    from data.api import ParamInfo, Request

    from pywikibot.site._apisite import _RequestWrapperT
    from pywikibot.site._namespace import NamespacesDict, SingleNamespaceType
    from pywikibot.site._tokenwallet import TokenWallet
    from pywikibot.tools import MediaWikiVersion


class GeneratorsMixin:

    """API generators mixin to MediaWiki site."""

    if typing.TYPE_CHECKING:
        _generator: Callable[..., _RequestWrapperT]
        _paraminfo: ParamInfo
        _request: Callable[..., Request]
        assert_valid_iter_params: Callable[..., None]
        encoding: Callable[[], str]
        get_property_names: Callable[..., list[str]]
        has_right: Callable[[str], bool]
        maxlimit: int
        mw_version: MediaWikiVersion
        namespaces: NamespacesDict
        protection_types: Callable[[], set[str]]
        sametitle: Callable[[str, str], bool]
        tokens: TokenWallet
        user: Callable[[], str | None]

    def load_pages_from_pageids(
        self,
        pageids: str | Iterable[int | str],
    ) -> Generator[pywikibot.Page, None, None]:
        """Return a page generator from pageids.

        Pages are iterated in the same order than in the underlying
        pageids.

        Pageids are filtered and only one page is returned in case of
        duplicate pageids.

        :param pageids: an iterable that returns pageids (str or int),
            or a comma- or pipe-separated string of pageids (e.g.
            '945097,1483753, 956608' or '945097|483753|956608')
        """
        if not pageids:
            return
        if isinstance(pageids, str):
            pageids = pageids.replace('|', ',')
            pageids = pageids.split(',')
            pageids = [p.strip() for p in pageids]

        # Validate pageids.
        gen = (str(int(p)) for p in pageids if int(p) > 0)

        for batch in batched(filter_unique(gen), self.maxlimit):
            # Store the order of the input data.
            priority_dict = dict(zip(batch, range(len(batch))))

            prio_queue: list[tuple[int, pywikibot.Page]] = []
            next_prio = 0
            params = {'pageids': batch}
            rvgen = api.PropertyGenerator('info', site=self, parameters=params)

            for pagedata in rvgen:
                title = pagedata['title']
                pageid = str(pagedata['pageid'])
                page = pywikibot.Page(pywikibot.Link(title, source=self))
                api.update_page(page, pagedata)
                priority, page = heapq.heappushpop(prio_queue,
                                                   (priority_dict[pageid],
                                                    page))
                # Smallest priority matches expected one; yield early.
                if priority == next_prio:
                    yield page
                    next_prio += 1
                else:
                    # Push onto the heap.
                    heapq.heappush(prio_queue, (priority, page))

            # Extract data in the same order of the input data.
            while prio_queue:
                priority, page = heapq.heappop(prio_queue)
                yield page

    def preloadpages(
        self,
        pagelist: Iterable[pywikibot.Page],
        *,
        groupsize: int | None = None,
        templates: bool = False,
        langlinks: bool = False,
        pageprops: bool = False,
        categories: bool = False,
        content: bool = True,
        quiet: bool = True,
    ) -> Generator[pywikibot.Page, None, None]:
        """Return a generator to a list of preloaded pages.

        Pages are iterated in the same order than in the underlying
        pagelist. In case of duplicates in a groupsize batch, return the
        first entry.

        .. versionchanged:: 7.6
           *content* parameter was added.
        .. versionchanged:: 7.7
           *categories* parameter was added.
        .. versionchanged:: 8.1
           *groupsize* is maxlimit by default. *quiet* parameter was
           added. No longer show the "Retrieving pages from site"
           message by default.

        :param pagelist: an iterable that returns Page objects
        :param groupsize: how many Pages to query at a time. If None
            (default), :attr:`maxlimit
            <pywikibot.site._apisite.APISite.maxlimit>` is used.
        :param templates: preload pages (typically templates)
            transcluded in the provided pages
        :param langlinks: preload all language links from the provided
            pages to other languages
        :param pageprops: preload various properties defined in page
            content
        :param categories: preload page categories
        :param content: preload page content
        :param quiet: If True (default), do not show the "Retrieving
            pages" message
        """
        props = 'revisions|info|categoryinfo'
        if templates:
            props += '|templates'
        if langlinks:
            props += '|langlinks'
        if pageprops:
            props += '|pageprops'
        if categories:
            props += '|categories'

        groupsize_ = min(groupsize or self.maxlimit, self.maxlimit)
        for batch in batched(pagelist, groupsize_):
            # Do not use p.pageid property as it will force page loading.
            pageids = [str(p._pageid) for p in batch
                       if hasattr(p, '_pageid') and p._pageid > 0]
            cache: dict[str, tuple[int, pywikibot.Page]] = {}
            # In case of duplicates, return the first entry.
            for priority, page in enumerate(batch):
                try:
                    cache.setdefault(page.title(with_section=False),
                                     (priority, page))
                except InvalidTitleError:
                    pywikibot.exception()

            prio_queue: list[tuple[int, pywikibot.Page]] = []
            next_prio = 0
            rvgen = api.PropertyGenerator(props, site=self)
            rvgen.set_maximum_items(-1)  # suppress use of "rvlimit" parameter

            if len(pageids) == len(batch) \
               and len(set(pageids)) <= self.maxlimit:
                # only use pageids if all pages have them
                rvgen.request['pageids'] = set(pageids)
            else:
                rvgen.request['titles'] = list(cache.keys())
            rvgen.request['rvprop'] = self._rvprops(content=content)
            if not quiet:
                pywikibot.info(f'Retrieving {len(cache)} pages from {self}.')

            for pagedata in rvgen:
                pywikibot.debug(f'Preloading {pagedata}')
                try:
                    if (pd_title := pagedata['title']) not in cache:
                        # API always returns a "normalized" title which is
                        # usually the same as the canonical form returned by
                        # page.title(), but sometimes not (e.g.,
                        # gender-specific localizations of "User" namespace).
                        # This checks to see if there is a normalized title in
                        # the response that corresponds to the canonical form
                        # used in the query.
                        for key, value in cache.items():
                            if self.sametitle(key, pd_title):
                                cache[pd_title] = value
                                break
                        else:
                            pywikibot.warning('preloadpages: Query returned '
                                              f'unexpected title {pd_title!r}')
                            continue

                except KeyError:
                    pywikibot.debug(f"No 'title' in {pagedata}\n"
                                    f'{pageids=!s}\n'
                                    f'titles={list(cache.keys())}')
                    continue

                priority, page = cache[pagedata['title']]
                api.update_page(page, pagedata, rvgen.props)
                priority, page = heapq.heappushpop(prio_queue,
                                                   (priority, page))
                # Smallest priority matches expected one; yield.
                if priority == next_prio:
                    yield page
                    next_prio += 1
                else:
                    # Push back onto the heap.
                    heapq.heappush(prio_queue, (priority, page))

            # Empty the heap.
            while prio_queue:
                priority, page = heapq.heappop(prio_queue)
                yield page

    def pagebacklinks(
        self,
        page: pywikibot.Page,
        *,
        follow_redirects: bool = False,
        filter_redirects=None,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate all pages that link to the given page.

        .. seealso:: :api:`Backlinks`

        :param page: The Page to get links to.
        :param follow_redirects: Also return links to redirects pointing to
            the given page.
        :param filter_redirects: If True, only return redirects to the given
            page. If False, only return non-redirect links. If None, return
            both (no filtering).
        :param namespaces: If present, only return links from the namespaces
            in this list.
        :param total: Maximum number of pages to retrieve in total.
        :param content: if True, load the current content of each iterated page
            (default False)
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        bltitle = page.title(with_section=False).encode(self.encoding())
        blargs: dict[str, Any] = {'gbltitle': bltitle}
        if filter_redirects is not None:
            blargs['gblfilterredir'] = ('redirects' if filter_redirects
                                        else 'nonredirects')
        blgen = self._generator(api.PageGenerator, type_arg='backlinks',
                                namespaces=namespaces, total=total,
                                g_content=content, **blargs)
        if follow_redirects:
            # links identified by MediaWiki as redirects may not really be,
            # so we have to check each "redirect" page and see if it
            # really redirects to this page
            # see fixed MediaWiki bug T9304
            redirgen = self._generator(api.PageGenerator,
                                       type_arg='backlinks',
                                       gbltitle=bltitle,
                                       gblfilterredir='redirects')
            genlist = {None: blgen}
            for redir in redirgen:
                if redir == page:
                    # if a wiki contains pages whose titles contain
                    # namespace aliases that existed before those aliases
                    # were defined (example: [[WP:Sandbox]] existed as a
                    # redirect to [[Wikipedia:Sandbox]] before the WP: alias
                    # was created) they can be returned as redirects to
                    # themselves; skip these
                    continue
                if redir.getRedirectTarget() == page:
                    genlist[redir.title()] = self.pagebacklinks(
                        redir, follow_redirects=True,
                        filter_redirects=filter_redirects,
                        namespaces=namespaces,
                        content=content
                    )
            return itertools.chain(*genlist.values())
        return blgen

    def page_embeddedin(
        self,
        page: pywikibot.Page,
        *,
        filter_redirects=None,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate all pages that embedded the given page as a template.

        .. seealso:: :api:`Embeddedin`

        :param page: The Page to get inclusions for.
        :param filter_redirects: If True, only return redirects that embed
            the given page. If False, only return non-redirect links. If
            None, return both (no filtering).
        :param namespaces: If present, only return links from the namespaces
            in this list.
        :param content: if True, load the current content of each iterated page
            (default False)
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        eiargs: dict[str, Any] = {
            'geititle': page.title(with_section=False).encode(self.encoding()),
        }
        if filter_redirects is not None:
            eiargs['geifilterredir'] = ('redirects' if filter_redirects
                                        else 'nonredirects')
        return self._generator(api.PageGenerator, type_arg='embeddedin',
                               namespaces=namespaces, total=total,
                               g_content=content, **eiargs)

    def page_redirects(
        self,
        page: pywikibot.Page,
        *,
        filter_fragments: bool | None = None,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
        content: bool = False
    ) -> Iterable[pywikibot.Page]:
        """Iterale all redirects to the given page.

        .. seealso:: :api:`Redirects`

        .. versionadded:: 7.0

        :param page: The Page to get redirects for.
        :param filter_fragments: If True, only return redirects with fragments.
            If False, only return redirects without fragments. If None, return
            both (no filtering).
        :param namespaces: Only return redirects from the namespaces
        :param total: maximum number of redirects to retrieve in total
        :param content: load the current content of each redirect
        """
        rdargs: dict[str, Any] = {
            'titles': page.title(with_section=False).encode(self.encoding()),
        }
        if filter_fragments is not None:
            rdargs['grdshow'] = ('' if filter_fragments else '!') + 'fragment'
        return self._generator(api.PageGenerator, type_arg='redirects',
                               namespaces=namespaces, total=total,
                               g_content=content, **rdargs)

    def pagereferences(
        self,
        page, *,
        follow_redirects: bool = False,
        filter_redirects: bool | None = None,
        with_template_inclusion: bool = True,
        only_template_inclusion: bool = False,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Convenience method combining pagebacklinks and page_embeddedin.

        :param namespaces: If present, only return links from the
            namespaces in this list.
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        if only_template_inclusion:
            return self.page_embeddedin(page,
                                        filter_redirects=filter_redirects,
                                        namespaces=namespaces, total=total,
                                        content=content)
        if not with_template_inclusion:
            return self.pagebacklinks(page, follow_redirects=follow_redirects,
                                      filter_redirects=filter_redirects,
                                      namespaces=namespaces, total=total,
                                      content=content)
        return itertools.islice(
            itertools.chain(
                self.pagebacklinks(
                    page, follow_redirects=follow_redirects,
                    filter_redirects=filter_redirects,
                    namespaces=namespaces, content=content),
                self.page_embeddedin(
                    page, filter_redirects=filter_redirects,
                    namespaces=namespaces, content=content)
            ), total)

    def pagelinks(
        self,
        page: pywikibot.page.BasePage,
        *,
        namespaces: NamespaceArgType = None,
        follow_redirects: bool = False,
        total: int | None = None,
        content: bool = False,
    ) -> Generator[pywikibot.Page, None, None]:
        """Yield internal wikilinks contained (or transcluded) on page.

        .. seealso::
           - :api:`Links`
           - :meth:`page.BasePage.linkedPages`

        :param namespaces: Only iterate pages in these namespaces
            (default: all)
        :param follow_redirects: if True, yields the target of any redirects,
            rather than the redirect page
        :param total: iterate no more than this number of pages in total
        :param content: if True, load the current content of each iterated page
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        plargs: dict[str, Any] = {}
        if hasattr(page, '_pageid'):
            plargs['pageids'] = str(page._pageid)
        else:
            pltitle = page.title(with_section=False).encode(self.encoding())
            plargs['titles'] = pltitle
        return self._generator(api.PageGenerator, type_arg='links',
                               namespaces=namespaces, total=total,
                               g_content=content, redirects=follow_redirects,
                               **plargs)

    def pagecategories(
        self,
        page: pywikibot.Page,
        *,
        with_sort_key: bool = False,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate categories to which page belongs.

        .. versionadded:: 9.6
           the *with_sort_key* parameter.

        .. seealso::
           - :meth:`page.BasePage.categories`
           - :api:`Categories`

        .. note:: This method also yields categories which are
           transcluded.

        :param with_sort_key: if True, include the sort key in each
            Category
        :param content: if True, load the current content of each
            iterated page default False); note that this means the
            contents of the category description page, not the pages
            contained in the category
        """
        clargs: dict[str, Any] = {}
        if hasattr(page, '_pageid'):
            clargs['pageids'] = str(page._pageid)
        else:
            clargs['titles'] = page.title(
                with_section=False).encode(self.encoding())

        if with_sort_key:
            page_dict = next(iter(self._generator(
                api.PropertyGenerator,
                type_arg='categories',
                total=total,
                clprop='sortkey',
                **clargs)))
            cats = (
                pywikibot.Category(self,
                                   cat_dict['title'],
                                   sort_key=cat_dict['sortkeyprefix'] or None)
                for cat_dict in page_dict.get('categories', [])
            )
            return self.preloadpages(cats) if content else cats

        return self._generator(api.PageGenerator,
                               type_arg='categories',
                               total=total,
                               g_content=content,
                               **clargs)

    def pageimages(
        self,
        page: pywikibot.Page,
        *,
        total: int | None = None,
        content: bool = False
    ) -> Iterable[pywikibot.Page]:
        """Iterate images used (not just linked) on the page.

        .. seealso:: :api:`Images`

        :param content: if True, load the current content of each iterated page
            (default False); note that this means the content of the image
            description page, not the image itself
        """
        imtitle = page.title(with_section=False).encode(self.encoding())
        return self._generator(api.PageGenerator, type_arg='images',
                               titles=imtitle, total=total,
                               g_content=content)

    def pagetemplates(
        self,
        page: pywikibot.Page,
        *,
        content: bool = False,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Iterate pages transcluded (not just linked) on the page.

        .. note:: You should not use this method directly; use
           :meth:`pywikibot.Page.itertemplates` instead.

        .. seealso::
           - :api:`Templates`
           - :meth:`page.BasePage.itertemplates`

        :param content: if True, load the current content of each
            iterated page (default False)
        :param namespaces: Only iterate pages in these namespaces
        :param total: maximum number of pages to retrieve in total

        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        :raises UnsupportedPageError: a Page object is not supported due
            to namespace restriction
        """
        tltitle = page.title(with_section=False).encode(self.encoding())
        return self._generator(api.PageGenerator, type_arg='templates',
                               titles=tltitle, namespaces=namespaces,
                               total=total, g_content=content)

    def categorymembers(
        self,
        category: pywikibot.Category,
        *,
        namespaces: NamespaceArgType = None,
        sortby: str | None = None,
        reverse: bool = False,
        starttime: pywikibot.time.Timestamp | None = None,
        endtime: pywikibot.time.Timestamp | None = None,
        total: int | None = None,
        startprefix: str | None = None,
        endprefix: str | None = None,
        content: bool = False,
        member_type: str | Iterable[str] | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Iterate members of specified category.

        You should not use this method directly; instead use one of the
        following:

        - :meth:`pywikibot.Category.articles`
        - :meth:`pywikibot.Category.members`
        - :meth:`pywikibot.Category.subcategories`

        .. versionchanged:: 4.0
           parameters except *category* are keyword arguments only.
        .. versionchanged:: 8.0
           raises TypeError instead of Error if no Category is specified
        .. seealso:: :api:`Categorymembers`

        :param category: The Category to iterate.
        :param namespaces: If present, only return category members from
            these namespaces. To yield subcategories or files, use
            parameter member_type instead.
        :param sortby: determines the order in which results are generated,
            valid values are "sortkey" (default, results ordered by category
            sort key) or "timestamp" (results ordered by time page was
            added to the category)
        :param reverse: if True, generate results in reverse order
            (default False)
        :param starttime: if provided, only generate pages added after this
            time; not valid unless sortby="timestamp"
        :param endtime: if provided, only generate pages added before this
            time; not valid unless sortby="timestamp"
        :param startprefix: if provided, only generate pages >= this title
            lexically; not valid if sortby="timestamp"
        :param endprefix: if provided, only generate pages < this title
            lexically; not valid if sortby="timestamp"
        :param content: if True, load the current content of each iterated page
            (default False)
        :param member_type: member type; values must be ``page``,
            ``subcat``, ``file``. If member_type includes ``page`` and
            is used in conjunction with sortby="timestamp", the API may
            limit results to only pages in the first 50 namespaces.

        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        :raises TypeError: no Category is specified
        :raises ValueError: invalid values given
        """
        if category.namespace() != 14:
            raise TypeError(
                f'categorymembers: non-Category page {category!r} specified')

        cmtitle = category.title(with_section=False).encode(self.encoding())
        cmargs: dict[str, Any] = {
            'type_arg': 'categorymembers',
            'gcmtitle': cmtitle,
            'gcmprop': 'ids|title|sortkey'
        }

        if sortby in ['sortkey', 'timestamp']:
            cmargs['gcmsort'] = sortby
        elif sortby:
            raise ValueError(
                f'categorymembers: invalid sortby value {sortby!r}')

        if starttime and endtime and starttime > endtime:
            raise ValueError(
                'categorymembers: starttime must be before endtime')

        if startprefix and endprefix and startprefix > endprefix:
            raise ValueError(
                'categorymembers: startprefix must be less than endprefix')

        if isinstance(member_type, str):
            member_type = {member_type}

        if member_type and sortby == 'timestamp':
            # Covert namespaces to a known type
            namespaces = set(self.namespaces.resolve(namespaces or []))

            if 'page' in member_type:
                excluded_namespaces = set()

                if 'file' not in member_type:
                    excluded_namespaces.add(6)
                if 'subcat' not in member_type:
                    excluded_namespaces.add(14)

                if namespaces:
                    if excluded_namespaces.intersection(namespaces):
                        raise ValueError(
                            f'incompatible namespaces {namespaces!r} and '
                            f'member_type {member_type!r}')
                    # All excluded namespaces are not present in `namespaces`.
                else:
                    # If the number of namespaces is greater than permitted by
                    # the API, it will issue a warning and use the namespaces
                    # up until the limit, which will usually be sufficient.
                    # TODO: QueryGenerator should detect when the number of
                    # namespaces requested is higher than available, and split
                    # the request into several batches.
                    excluded_namespaces.update([-1, -2])
                    namespaces = set(self.namespaces) - excluded_namespaces
            else:
                if 'file' in member_type:
                    namespaces.add(6)
                if 'subcat' in member_type:
                    namespaces.add(14)

            member_type = None

        if member_type:
            cmargs['gcmtype'] = member_type

        if reverse:
            cmargs['gcmdir'] = 'desc'
            # API wants start/end params in opposite order if using descending
            # sort; we take care of this reversal for the user
            starttime, endtime = endtime, starttime
            startprefix, endprefix = endprefix, startprefix

        if starttime and sortby == 'timestamp':
            cmargs['gcmstart'] = starttime
        elif starttime:
            raise ValueError('categorymembers: '
                             "invalid combination of 'sortby' and 'starttime'")

        if endtime and sortby == 'timestamp':
            cmargs['gcmend'] = endtime
        elif endtime:
            raise ValueError('categorymembers: '
                             "invalid combination of 'sortby' and 'endtime'")

        if startprefix and sortby != 'timestamp':
            cmargs['gcmstartsortkeyprefix'] = startprefix
        elif startprefix:
            raise ValueError('categorymembers: invalid combination of '
                             "'sortby' and 'startprefix'")

        if endprefix and sortby != 'timestamp':
            cmargs['gcmendsortkeyprefix'] = endprefix
        elif endprefix:
            raise ValueError('categorymembers: '
                             "invalid combination of 'sortby' and 'endprefix'")

        return self._generator(api.PageGenerator, namespaces=namespaces,
                               total=total, g_content=content, **cmargs)

    def _rvprops(self, content: bool = False) -> list[str]:
        """Setup rvprop items for loadrevisions and preloadpages.

        :return: rvprop items
        """
        props = ['comment', 'contentmodel', 'flags', 'ids', 'parsedcomment',
                 'sha1', 'size', 'tags', 'timestamp', 'user', 'userid']
        if content:
            props.append('content')
        if self.mw_version >= '1.32':
            props.append('roles')
        return props

    def loadrevisions(
        self,
        page: pywikibot.Page,
        *,
        content: bool = False,
        section: int | None = None,
        **kwargs,
    ) -> None:
        """Retrieve revision information and store it in page object.

        By default, retrieves the last (current) revision of the page,
        unless any of the optional parameters revids, startid, endid,
        starttime, endtime, rvdir, user, excludeuser, or total are
        specified. Unless noted below, all parameters not specified
        default to False.

        If rvdir is False or not specified, startid must be greater than
        endid if both are specified; likewise, starttime must be greater
        than endtime. If rvdir is True, these relationships are reversed.

        .. seealso:: :api:`Revisions`

        :param page: retrieve revisions of this Page and hold the data.
        :param content: if True, retrieve the wiki-text of each revision;
            otherwise, only retrieve the revision metadata (default)
        :param section: if specified, retrieve only this section of the text
            (content must be True); section must be given by number (top of
            the article is section 0), not name
        :keyword revids: retrieve only the specified revision ids (raise
            Exception if any of revids does not correspond to page)
        :type revids: an int, a str or a list of ints or strings
        :keyword startid: retrieve revisions starting with this revid
        :keyword endid: stop upon retrieving this revid
        :keyword starttime: retrieve revisions starting at this Timestamp
        :keyword endtime: stop upon reaching this Timestamp
        :keyword rvdir: if false, retrieve newest revisions first (default);
            if true, retrieve oldest first
        :keyword user: retrieve only revisions authored by this user
        :keyword excludeuser: retrieve all revisions not authored by this user
        :keyword total: number of revisions to retrieve
        :raises ValueError: invalid startid/endid or starttime/endtime values
        :raises pywikibot.exceptions.Error: revids belonging to a different
            page
        """
        latest = all(val is None for val in kwargs.values())

        revids = kwargs.get('revids')
        startid = kwargs.get('startid')
        starttime = kwargs.get('starttime')
        endid = kwargs.get('endid')
        endtime = kwargs.get('endtime')
        rvdir = kwargs.get('rvdir')
        user = kwargs.get('user')
        step = kwargs.get('step')

        # check for invalid argument combinations
        if (startid is not None or endid is not None) \
           and (starttime is not None or endtime is not None):
            raise ValueError(
                'loadrevisions: startid/endid combined with starttime/endtime')

        if starttime is not None and endtime is not None:
            if rvdir and starttime >= endtime:
                raise ValueError(
                    'loadrevisions: starttime > endtime with rvdir=True')

            if not rvdir and endtime >= starttime:
                raise ValueError(
                    'loadrevisions: endtime > starttime with rvdir=False')

        if startid is not None and endid is not None:
            if rvdir and startid >= endid:
                raise ValueError(
                    'loadrevisions: startid > endid with rvdir=True')
            if not rvdir and endid >= startid:
                raise ValueError(
                    'loadrevisions: endid > startid with rvdir=False')

        rvargs: dict[str, Any] = {
            'type_arg': 'info|revisions',
            'rvprop': self._rvprops(content=content),
        }

        if content and section is not None:
            rvargs['rvsection'] = str(section)

        if revids is None:
            rvtitle = page.title(with_section=False).encode(self.encoding())
            rvargs['titles'] = rvtitle
        else:
            if isinstance(revids, (int, str)):
                ids = str(revids)
            else:
                ids = '|'.join(str(r) for r in revids)
            rvargs['revids'] = ids

        if rvdir:
            rvargs['rvdir'] = 'newer'
        elif rvdir is not None:
            rvargs['rvdir'] = 'older'

        if startid:
            rvargs['rvstartid'] = startid
        if endid:
            rvargs['rvendid'] = endid
        if starttime:
            rvargs['rvstart'] = starttime
        if endtime:
            rvargs['rvend'] = endtime

        if user:
            rvargs['rvuser'] = user
        else:
            rvargs['rvexcludeuser'] = kwargs.get('excludeuser')

        # assemble API request
        rvgen = self._generator(api.PropertyGenerator,
                                total=kwargs.get('total'), **rvargs)

        if step:
            rvgen.set_query_increment = step

        if latest or 'revids' in rvgen.request:
            rvgen.set_maximum_items(-1)  # suppress use of rvlimit parameter

        for pagedata in rvgen:
            if not self.sametitle(pagedata['title'],
                                  page.title(with_section=False)):
                raise InconsistentTitleError(page, pagedata['title'])
            if 'missing' in pagedata:
                raise NoPageError(page)
            api.update_page(page, pagedata, rvgen.props)

    def pagelanglinks(
        self,
        page: pywikibot.Page,
        *,
        total: int | None = None,
        include_obsolete: bool = False,
        include_empty_titles: bool = False,
    ) -> Generator[pywikibot.Link, None, None]:
        """Yield all interlanguage links on page, yielding Link objects.

        .. versionchanged:: 6.2:
           `include_empty_titles` parameter was added.

        .. seealso:: :api:`Langlinks`

        :param include_obsolete: if true, yield even Link objects whose
            site is obsolete
        :param include_empty_titles: if true, yield even Link objects whose
            title is empty but redirects to a site like [[en:]]
        """
        lltitle = page.title(with_section=False)
        llquery = self._generator(api.PropertyGenerator,
                                  type_arg='langlinks',
                                  titles=lltitle.encode(self.encoding()),
                                  total=total)
        for pageitem in llquery:
            if not self.sametitle(pageitem['title'], lltitle):
                raise InconsistentTitleError(page, pageitem['title'])
            if 'langlinks' not in pageitem:
                continue
            for linkdata in pageitem['langlinks']:
                link = pywikibot.Link.langlinkUnsafe(linkdata['lang'],
                                                     linkdata['*'],
                                                     source=self)
                if link.site.obsolete and not include_obsolete:
                    continue

                if link.title or include_empty_titles:
                    yield link

    def page_extlinks(
        self,
        page: pywikibot.Page,
        *,
        total: int | None = None,
    ) -> Generator[str, None, None]:
        """Yield all external links on page, yielding URL strings.

        .. seealso:: :api:`Extlinks`
        """
        eltitle = page.title(with_section=False)
        elquery = self._generator(api.PropertyGenerator, type_arg='extlinks',
                                  titles=eltitle.encode(self.encoding()),
                                  total=total)
        for pageitem in elquery:
            if not self.sametitle(pageitem['title'], eltitle):
                raise InconsistentTitleError(page, pageitem['title'])
            if 'extlinks' not in pageitem:
                continue
            for linkdata in pageitem['extlinks']:
                yield linkdata['*']

    @deprecated_signature(since='10.4.0')
    def allpages(
        self,
        start: str = '!', *,
        prefix: str = '',
        namespace: SingleNamespaceType = 0,
        filterredir: bool | None = None,
        filterlanglinks: bool | None = None,
        minsize: int | None = None,
        maxsize: int | None = None,
        protect_type: str | None = None,
        protect_level: str | None = None,
        reverse: bool = False,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate pages in a single namespace.

        .. seealso:: :api:`Allpages`

        :param start: Start at this title (page need not exist).
        :param prefix: Only yield pages starting with this string.
        :param namespace: Iterate pages from this (single) namespace.
        :param filterredir: if True, only yield redirects; if False (and
            not None), only yield non-redirects (default: yield both).
        :param filterlanglinks: if True, only yield pages with language
            links; if False (and not None), only yield pages without
            language links (default: yield both).
        :param minsize: if present, only yield pages at least this many
            bytes in size.
        :param maxsize: if present, only yield pages at most this many
            bytes in size.
        :param protect_type: only yield pages that have a protection of
            the specified type.
        :param protect_level: only yield pages that have protection at
            this level; can only be used if protect_type is specified.
        :param reverse: if True, iterate in reverse Unicode lexicographic
            order (default: iterate in forward order).
        :param content: if True, load the current content of each
            iterated page (default False).
        :raises KeyError: the namespace identifier was not resolved
        :raises TypeError: the namespace identifier has an inappropriate
            type such as bool, or an iterable with more than one
            namespace or *filterredir* parameter has an invalid type.
        """
        def _maxsize_filter(item):
            """Return True if page text length is within maxsize limit."""
            return len(item.text.encode(self.encoding())) <= maxsize

        misermode = self.siteinfo.get('misermode') and maxsize is not None
        if filterredir not in (True, False, None):
            raise TypeError('filterredir parameter must be True, False or '
                            f'None, not {type(filterredir)}')

        apgen = self._generator(api.PageGenerator, type_arg='allpages',
                                namespaces=namespace,
                                gapfrom=start, total=total,
                                g_content=content or misermode)
        if prefix:
            apgen.request['gapprefix'] = prefix
        if filterredir is not None:
            apgen.request['gapfilterredir'] = ('redirects' if filterredir else
                                               'nonredirects')
        if filterlanglinks is not None:
            apgen.request['gapfilterlanglinks'] = ('withlanglinks'
                                                   if filterlanglinks else
                                                   'withoutlanglinks')
        if isinstance(minsize, int):
            apgen.request['gapminsize'] = str(minsize)
        if not misermode and isinstance(maxsize, int):
            apgen.request['gapmaxsize'] = str(maxsize)
        if isinstance(protect_type, str):
            apgen.request['gapprtype'] = protect_type
            if isinstance(protect_level, str):
                apgen.request['gapprlevel'] = protect_level
        if reverse:
            apgen.request['gapdir'] = 'descending'
        if misermode:
            apgen.filter_func = _maxsize_filter

        return apgen

    @deprecated(since='10.7.0')
    def alllinks(
        self,
        start: str = '',
        prefix: str = '',
        namespace: SingleNamespaceType = 0,
        unique: bool = False,
        fromids: bool = False,
        total: int | None = None,
    ) -> Generator[pywikibot.Page, None, None]:
        """Iterate all links to pages (which need not exist) in one namespace.

        .. note:: In practice, links that were found on pages that have
           been deleted may not have been removed from the links table,
           so this method can return false positives.

        .. caution:: *unique* parameter is no longer supported by
           MediaWiki 1.43 or higher. Pywikibot uses
           :func:`tools.itertools.filter_unique` in that case which
           might be memory intensive. Use it with care.

        .. important:: Using *namespace* option different from ``0``
           needs a lot of time on Wikidata site. You have to increase
           the **read** timeout part of ``socket_timeout`` in
           :ref:`Http Settings` in your ``user-config.py`` file. Or
           increase it partially within your code like:

           .. code-block:: python

              from pywikibot import config
              save_timeout = config.socket_timeout  # save the timeout config
              config.socket_timeout = save_timeout[0], 60
              ... # your code here
              config.socket_timeout = save_timeout  # restore timeout config

           The minimum read timeout value should be 60 seconds in that
           case.

        .. deprecated:: 10.7
           This method is dysfunctional and should no longer be used. It
           will probably be removed in Pywikibot 11.

        .. seealso::
           - :api:`Alllinks`
           - :meth:`pagebacklinks`
           - :meth:`pagelinks`

        :param start: Start at this title (page need not exist).
        :param prefix: Only yield pages starting with this string.
        :param namespace: Iterate pages from this (single) namespace
        :param unique: If True, only iterate each link title once
            (default: False)
        :param fromids: if True, include the pageid of the page
            containing each link (default: False) as the '_fromid'
            attribute of the Page; cannot be combined with *unique*
        :raises KeyError: the *namespace* identifier was not resolved
        :raises TypeError: the *namespace* identifier has an
            inappropriate type such as bool, or an iterable with more
            than one namespace
        """
        # no cover: start
        if unique and fromids:
            raise Error('alllinks: unique and fromids cannot both be True.')
        algen = self._generator(api.ListGenerator, type_arg='alllinks',
                                namespaces=namespace, total=total)
        if fromids:
            algen.request['alprop'] = 'title|ids'

        # circumvent problems with unique and prefix due to T359425 and T359427
        if self.mw_version < '1.43':
            if prefix:
                algen.request['alprefix'] = prefix
                prefix = ''  # don't break the loop later
            if start:
                algen.request['alfrom'] = start
            algen.request['alunique'] = unique
        else:
            if prefix:
                algen.request['alfrom'] = prefix
            elif start:
                algen.request['alfrom'] = start
            if unique:
                algen = filter_unique(
                    algen, key=lambda link: (link['title'], link['ns']))

        for link in algen:
            p = pywikibot.Page(self, link['title'], link['ns'])
            if prefix and p.title() > prefix:  # T359425, T359427
                break
            if fromids:
                p._fromid = link['fromid']  # type: ignore[attr-defined]
            yield p
        # no cover: stop

    def allcategories(
        self,
        start: str = '!',
        prefix: str = '',
        total: int | None = None,
        reverse: bool = False,
        content: bool = False,
    ) -> Iterable[pywikibot.page.Category]:
        """Iterate categories used (which need not have a Category page).

        Iterator yields Category objects. Note that, in practice, links that
        were found on pages that have been deleted may not have been removed
        from the database table, so this method can return false positives.

        .. seealso:: :api:`Allcategories`

        :param start: Start at this category title (category need not exist).
        :param prefix: Only yield categories starting with this string.
        :param reverse: if True, iterate in reverse Unicode lexicographic
            order (default: iterate in forward order)
        :param content: if True, load the current content of each iterated page
            (default False); note that this means the contents of the category
            description page, not the pages that are members of the category
        """
        acgen = self._generator(api.PageGenerator,
                                type_arg='allcategories', gacfrom=start,
                                total=total, g_content=content)
        if prefix:
            acgen.request['gacprefix'] = prefix
        if reverse:
            acgen.request['gacdir'] = 'descending'
        return acgen

    def botusers(
        self,
        total: int | None = None,
    ) -> Generator[dict[str, Any], None, None]:
        """Iterate bot users.

        Iterated values are dicts containing 'name', 'userid',
        'editcount', 'registration', and 'groups' keys. 'groups' will be
        present only if the user is a member of at least 1 group, and
        will be a list of str; all the other values are str and should
        always be present.
        """
        if not hasattr(self, '_bots'):
            self._bots: dict[str, dict[str, str | list[str]]] = {}

        if not self._bots:
            for item in self.allusers(group='bot', total=total):
                assert isinstance(item['name'], str)
                self._bots.setdefault(item['name'], item)

        yield from self._bots.values()

    def allusers(
        self,
        start: str = '!',
        prefix: str = '',
        group: str | None = None,
        total: int | None = None,
    ) -> Iterable[dict[str, str | list[str]]]:
        """Iterate registered users, ordered by username.

        Iterated values are dicts containing 'name', 'editcount',
        'registration', and (sometimes) 'groups' keys. 'groups' will be
        present only if the user is a member of at least 1 group, and
        will be a list of str; all the other values are str and should
        always be present.

        .. seealso:: :api:`Allusers`

        :param start: start at this username (name need not exist)
        :param prefix: only iterate usernames starting with this substring
        :param group: only iterate users that are members of this group
        """
        augen = self._generator(api.ListGenerator, type_arg='allusers',
                                auprop='editcount|groups|registration',
                                aufrom=start, total=total)
        if prefix:
            augen.request['auprefix'] = prefix
        if group:
            augen.request['augroup'] = group
        return augen

    def allimages(
        self,
        start: str = '!',
        prefix: str = '',
        minsize: int | None = None,
        maxsize: int | None = None,
        reverse: bool = False,
        sha1: str | None = None,
        sha1base36: str | None = None,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.FilePage]:
        """Iterate all images, ordered by image title.

        Yields FilePages, but these pages need not exist on the wiki.

        .. seealso:: :api:`Allimages`

        :param start: start at this title (name need not exist)
        :param prefix: only iterate titles starting with this substring
        :param minsize: only iterate images of at least this many bytes
        :param maxsize: only iterate images of no more than this many bytes
        :param reverse: if True, iterate in reverse lexicographic order
        :param sha1: only iterate image (it is theoretically possible there
            could be more than one) with this sha1 hash
        :param sha1base36: same as sha1 but in base 36
        :param content: if True, load the current content of each iterated page
            (default False); note that this means the content of the image
            description page, not the image itself
        """
        aigen = self._generator(api.PageGenerator,
                                type_arg='allimages', gaifrom=start,
                                total=total, g_content=content)
        if prefix:
            aigen.request['gaiprefix'] = prefix
        if isinstance(minsize, int):
            aigen.request['gaiminsize'] = str(minsize)
        if isinstance(maxsize, int):
            aigen.request['gaimaxsize'] = str(maxsize)
        if reverse:
            aigen.request['gaidir'] = 'descending'
        if sha1:
            aigen.request['gaisha1'] = sha1
        if sha1base36:
            aigen.request['gaisha1base36'] = sha1base36
        return aigen

    def filearchive(
        self,
        start: str | None = None,
        end: str | None = None,
        reverse: bool = False,
        total: int | None = None,
        **kwargs,
    ) -> Iterable[dict[str, Any]]:
        """Iterate archived files.

        Yields dict of file archive information.

        .. seealso:: :api:`filearchive`

        :param start: start at this title (name need not exist)
        :param end: end at this title (name need not exist)
        :param reverse: if True, iterate in reverse lexicographic order
        :param total: maximum number of pages to retrieve in total
        :keyword prefix: only iterate titles starting with this substring
        :keyword sha1: only iterate image with this sha1 hash
        :keyword sha1base36: same as sha1 but in base 36
        :keyword prop: Image information to get. Default is timestamp
        """
        if start and end:
            self.assert_valid_iter_params(
                'filearchive', start, end, reverse, is_ts=False)
        fagen = self._generator(api.ListGenerator,
                                type_arg='filearchive',
                                fafrom=start,
                                fato=end,
                                total=total)
        for k, v in kwargs.items():
            fagen.request['fa' + k] = v
        if reverse:
            fagen.request['fadir'] = 'descending'
        return fagen

    def blocks(
        self,
        starttime: pywikibot.time.Timestamp | None = None,
        endtime: pywikibot.time.Timestamp | None = None,
        reverse: bool = False,
        blockids: int | str | Iterable[int | str] | None = None,
        users: str | Iterable[str] | None = None,
        iprange: str | None = None,
        total: int | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Iterate all current blocks, in order of creation.

        The iterator yields dicts containing keys corresponding to the
        block properties.

        .. seealso:: :api:`Blocks`

        .. note:: logevents only logs user blocks,
           while this method iterates all blocks
           including IP ranges.
        .. warning::
           ``iprange`` parameter cannot be used together with ``users``.

        :param starttime: start iterating at this Timestamp
        :param endtime: stop iterating at this Timestamp
        :param reverse: if True, iterate oldest blocks first (default: newest)
        :param blockids: only iterate blocks with these id numbers. Numbers
            must be separated by '|' if given by a str.
        :param users: only iterate blocks affecting these usernames or IPs
        :param iprange: a single IP or an IP range. Ranges broader than
            IPv4/16 or IPv6/19 are not accepted.
        :param total: total amount of block entries
        """
        if starttime and endtime:
            self.assert_valid_iter_params('blocks', starttime, endtime,
                                          reverse)
        bkgen = self._generator(api.ListGenerator, type_arg='blocks',
                                total=total)
        bkgen.request['bkprop'] = ['id', 'user', 'by', 'timestamp', 'expiry',
                                   'reason', 'range', 'flags', 'userid']
        if starttime:
            bkgen.request['bkstart'] = starttime
        if endtime:
            bkgen.request['bkend'] = endtime
        if reverse:
            bkgen.request['bkdir'] = 'newer'
        if blockids:
            bkgen.request['bkids'] = blockids
        if users:
            if isinstance(users, str):
                users = users.split('|')

            # actual IPv6 addresses (anonymous users) are uppercase, but they
            # have never a :: in the username (so those are registered users)
            users = [user.upper() if is_ip_address(user) and '::' not in user
                     else user for user in users]
            bkgen.request['bkusers'] = users
        elif iprange:
            bkgen.request['bkip'] = iprange
        return bkgen

    def exturlusage(
        self,
        url: str | None = None,
        protocol: str | None = None,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate Pages that contain links to the given URL.

        .. seealso:: :api:`Exturlusage`

        :param url: The URL to search for (with or without the protocol
            prefix); this may include a '*' as a wildcard, only at the start
            of the hostname
        :param namespaces: list of namespace numbers to fetch contribs from
        :param total: Maximum number of pages to retrieve in total
        :param protocol: protocol to search for, http and https by default.
                Full list shown on Special:LinkSearch wikipage
        """
        if url is not None:
            found_protocol, _, url = url.rpartition('://')

            # If url is * we make it None in order to search for every page
            # with any URL.
            if url == '*':
                url = None

            if found_protocol:
                if protocol and protocol != found_protocol:
                    raise ValueError(
                        f'Protocol {protocol!r} was specified, but '
                        f'{found_protocol!r} was found in searched url'
                    )
                protocol = found_protocol

        return self._generator(api.PageGenerator, type_arg='exturlusage',
                               geuquery=url, geuprotocol=protocol,
                               namespaces=namespaces,
                               total=total, g_content=content)

    def imageusage(
        self,
        image: pywikibot.FilePage,
        *,
        namespaces: NamespaceArgType = None,
        filterredir: bool | None = None,
        total: int | None = None,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate Pages that contain links to the given FilePage.

        .. seealso:: :api:`Imageusage`
        .. versionchanged:: 7.2
           all parameters except `image` are keyword only.

        :param image: the image to search for (FilePage need not exist on
            the wiki)
        :param namespaces: If present, only iterate pages in these namespaces
        :param filterredir: if True, only yield redirects; if False (and not
            None), only yield non-redirects (default: yield both)
        :param total: iterate no more than this number of pages in total
        :param content: if True, load the current content of each iterated page
            (default False)
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        iuargs = {'giutitle': image.title(with_section=False)}
        if filterredir is not None:
            iuargs['giufilterredir'] = ('redirects' if filterredir else
                                        'nonredirects')
        return self._generator(api.PageGenerator, type_arg='imageusage',
                               namespaces=namespaces,
                               total=total, g_content=content, **iuargs)

    def logevents(
        self,
        logtype: str | None = None,
        user: str | None = None,
        page: str | pywikibot.Page | None = None,
        namespace: NamespaceArgType = None,
        start: str | pywikibot.time.Timestamp | None = None,
        end: str | pywikibot.time.Timestamp | None = None,
        reverse: bool = False,
        tag: str | None = None,
        total: int | None = None,
    ) -> Iterable[pywikibot.logentries.LogEntry]:
        """Iterate all log entries.

        .. seealso:: :api:`Logevents`

        .. note:: logevents with `logtype='block'` only logs user blocks
           whereas `site.blocks` iterates all blocks including IP ranges.

        :param logtype: only iterate entries of this type
            (see mediawiki api documentation for available types)
        :param user: only iterate entries that match this user name
        :param page: only iterate entries affecting this page
        :param namespace: namespace(s) to retrieve logevents from

        .. note:: due to an API limitation,
           if namespace param contains multiple namespaces,
           log entries from all namespaces will be fetched from
           the API and will be filtered later during iteration.

        :param start: only iterate entries from and after this Timestamp
        :param end: only iterate entries up to and through this Timestamp
        :param reverse: if True, iterate oldest entries first (default: newest)
        :param tag: only iterate entries tagged with this tag
        :param total: maximum number of events to iterate

        :raises KeyError: the namespace identifier was not resolved
        :raises TypeError: the namespace identifier has an inappropriate
            type such as bool, or an iterable with more than one namespace
        """
        if start and end:
            self.assert_valid_iter_params('logevents', start, end, reverse)

        legen = self._generator(api.LogEntryListGenerator, type_arg=logtype,
                                total=total)
        if logtype is not None:
            legen.request['letype'] = logtype
        if user is not None:
            legen.request['leuser'] = user
        if page is not None:
            legen.request['letitle'] = page
        if start is not None:
            legen.request['lestart'] = start
        if end is not None:
            legen.request['leend'] = end
        if reverse:
            legen.request['ledir'] = 'newer'
        if namespace is not None:
            legen.set_namespace(namespace)
        if tag:
            legen.request['letag'] = tag

        return legen

    def recentchanges(
        self,
        *,
        start: pywikibot.time.Timestamp | None = None,
        end: pywikibot.time.Timestamp | None = None,
        reverse: bool = False,
        namespaces: NamespaceArgType = None,
        changetype: str | None = None,
        minor: bool | None = None,
        bot: bool | None = None,
        anon: bool | None = None,
        redirect: bool | None = None,
        patrolled: bool | None = None,
        top_only: bool = False,
        page: str | pywikibot.Page | None = None,
        total: int | None = None,
        user: str | None = None,
        excludeuser: str | None = None,
        tag: str | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Iterate recent changes.

        .. versionchanged:: 10.1
           *page* parameter was added.

        .. seealso:: :api:`RecentChanges`

        :param start: Timestamp to start listing from
        :type start: pywikibot.Timestamp
        :param end: Timestamp to end listing at
        :type end: pywikibot.Timestamp
        :param reverse: if True, start with oldest changes (default: newest)
        :param namespaces: only iterate pages in these namespaces
        :param changetype: only iterate changes of this type ("edit" for
            edits to existing pages, "new" for new pages, "log" for log
            entries)
        :param minor: if True, only list minor edits; if False, only list
            non-minor edits; if None, list all
        :param bot: if True, only list bot edits; if False, only list
            non-bot edits; if None, list all
        :param anon: if True, only list anon edits; if False, only list
            non-anon edits; if None, list all
        :param redirect: if True, only list edits to redirect pages; if
            False, only list edits to non-redirect pages; if None, list all
        :param patrolled: if True, only list patrolled edits; if False,
            only list non-patrolled edits; if None, list all
        :param top_only: if True, only list changes that are the latest
            revision (default False)
        :param page: if not None, only list edits to this page
        :param user: if not None, only list edits by this user
        :param excludeuser: if not None, exclude edits by this user
        :param tag: a recent changes tag
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        if start and end:
            self.assert_valid_iter_params('recentchanges', start, end, reverse)

        rcgen = self._generator(api.ListGenerator, type_arg='recentchanges',
                                rcprop='user|comment|timestamp|title|ids'
                                       '|sizes|redirect|loginfo|flags|tags',
                                namespaces=namespaces,
                                total=total, rctoponly=top_only)
        if start is not None:
            rcgen.request['rcstart'] = start
        if end is not None:
            rcgen.request['rcend'] = end
        if reverse:
            rcgen.request['rcdir'] = 'newer'
        if changetype:
            rcgen.request['rctype'] = changetype
        if page:
            rcgen.request['rctitle'] = page
        filters = {'minor': minor,
                   'bot': bot,
                   'anon': anon,
                   'redirect': redirect,
                   }
        if patrolled is not None and (
                self.has_right('patrol') or self.has_right('patrolmarks')):
            rcgen.request['rcprop'] += ['patrolled']
            filters['patrolled'] = patrolled
        rcgen.request['rcshow'] = api.OptionSet(self, 'recentchanges', 'show',
                                                filters)

        if user:
            rcgen.request['rcuser'] = user

        if excludeuser:
            rcgen.request['rcexcludeuser'] = excludeuser
        rcgen.request['rctag'] = tag
        return rcgen

    def search(
        self,
        searchstring: str,
        *,
        namespaces: NamespaceArgType = None,
        where: typing.Literal['text', 'title', 'nearmatch'] | None = None,
        total: int | None = None,
        content: bool = False,
        sort: typing.Literal[
            'create_timestamp_asc', 'create_timestamp_desc',
            'incoming_links_asc', 'incoming_links_desc', 'just_match',
            'last_edit_asc', 'last_edit_desc', 'none', 'random', 'relevance',
            'user_random'
        ] | None = 'relevance',
    ) -> Iterable[pywikibot.Page]:
        """Iterate Pages that contain the searchstring.

        Note that this may include non-existing Pages if the wiki's
        database table contains outdated entries.

        .. versionchanged:: 7.0
           Default of `where` parameter has been changed from 'text' to
           None. The behaviour depends on the installed search engine
           which is 'text' on CirrusSearch'.
           raises APIError instead of Error if searchstring is not set
           or what parameter is wrong.
        .. versionchanged:: 10.0
           The *sort* parameter was added.

        .. seealso:: :api:`Search`

        :param searchstring: the text to search for
        :param namespaces: search only in these namespaces (defaults to
            all)
        :param where: Where to search; value must be one of the given
            literals or None (many wikis do not support all search types)
        :param total: limit result to this number of pages
        :param content: if True, load the current content of each
            iterated page (default False)
        :param sort: Set the sort order of returned results. If None is
            given, 'none' is used. Default is sort by relevance.
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        :raises APIError: The "gsrsearch" parameter must be set:
            searchstring parameter is not set
        :raises APIError: Unrecognized value for parameter "gsrwhat":
            wrong where parameter is given
        """
        if not namespaces and namespaces != 0:
            namespaces = [ns_id for ns_id in self.namespaces if ns_id >= 0]
        parameters = {
            'gsrsearch': searchstring,
            'gsrwhat': where,
        }
        if sort is None:
            sort = 'none'
        if sort != 'relevance':
            parameters['gsrsort'] = sort

        return self._generator(api.PageGenerator, type_arg='search',
                               namespaces=namespaces,
                               total=total, g_content=content,
                               parameters=parameters)

    def usercontribs(
        self,
        user: str | None = None,
        userprefix: str | None = None,
        start=None,
        end=None,
        reverse: bool = False,
        namespaces: NamespaceArgType = None,
        minor: bool | None = None,
        total: int | None = None,
        top_only: bool = False,
    ) -> Iterable[dict[str, Any]]:
        """Iterate contributions by a particular user.

        Iterated values are in the same format as recentchanges.

        .. seealso::
           - :api:`Usercontribs`
           - :meth:`pywikibot.User.contributions`

        :param user: Iterate contributions by this user (name or IP)
        :param userprefix: Iterate contributions by all users whose names
            or IPs start with this substring
        :param start: Iterate contributions starting at this Timestamp
        :param end: Iterate contributions ending at this Timestamp
        :param reverse: Iterate oldest contributions first (default: newest)
        :param namespaces: only iterate pages in these namespaces
        :param minor: if True, iterate only minor edits; if False and
            not None, iterate only non-minor edits (default: iterate both)
        :param total: limit result to this number of pages
        :param top_only: if True, iterate only edits which are the latest
            revision (default: False)
        :raises pywikibot.exceptions.Error: either user or userprefix must be
            non-empty
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        if not (user or userprefix):
            raise Error(
                'usercontribs: either user or userprefix must be non-empty')

        if start and end:
            self.assert_valid_iter_params('usercontribs', start, end, reverse)

        ucgen = self._generator(api.ListGenerator, type_arg='usercontribs',
                                ucprop='ids|title|timestamp|comment|flags',
                                namespaces=namespaces,
                                total=total, uctoponly=top_only)
        if user:
            ucgen.request['ucuser'] = user
        if userprefix:
            ucgen.request['ucuserprefix'] = userprefix
        if start is not None:
            ucgen.request['ucstart'] = str(start)
        if end is not None:
            ucgen.request['ucend'] = str(end)
        if reverse:
            ucgen.request['ucdir'] = 'newer'
        option_set = api.OptionSet(self, 'usercontribs', 'show')
        option_set['minor'] = minor
        ucgen.request['ucshow'] = option_set
        return ucgen

    def watchlist_revs(
        self,
        start=None,
        end=None,
        reverse: bool = False,
        namespaces: NamespaceArgType = None,
        minor: bool | None = None,
        bot: bool | None = None,
        anon: bool | None = None,
        total: int | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Iterate revisions to pages on the bot user's watchlist.

        Iterated values will be in same format as recentchanges.

        .. seealso:: :api:`Watchlist`

        :param start: Iterate revisions starting at this Timestamp
        :param end: Iterate revisions ending at this Timestamp
        :param reverse: Iterate oldest revisions first (default: newest)
        :param namespaces: only iterate pages in these namespaces
        :param minor: if True, only list minor edits; if False (and not
            None), only list non-minor edits
        :param bot: if True, only list bot edits; if False (and not
            None), only list non-bot edits
        :param anon: if True, only list anon edits; if False (and not
            None), only list non-anon edits
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        if start and end:
            self.assert_valid_iter_params(
                'watchlist_revs', start, end, reverse)

        wlgen = self._generator(
            api.ListGenerator, type_arg='watchlist',
            wlprop='user|comment|timestamp|title|ids|flags',
            wlallrev='', namespaces=namespaces, total=total)
        # TODO: allow users to ask for "patrol" as well?
        if start is not None:
            wlgen.request['wlstart'] = start
        if end is not None:
            wlgen.request['wlend'] = end
        if reverse:
            wlgen.request['wldir'] = 'newer'
        filters = {'minor': minor, 'bot': bot, 'anon': anon}
        wlgen.request['wlshow'] = api.OptionSet(self, 'watchlist', 'show',
                                                filters)
        return wlgen

    def _check_view_deleted(self, msg_prefix: str, prop: list[str]) -> None:
        """Check if the user can view deleted comments and content.

        :param msg_prefix: The calling method name
        :param prop: Requested props to check
        :raises UserRightsError: user cannot view a requested prop
        """
        err = f'{msg_prefix}: User:{self.user()} not authorized to view '
        if not self.has_right('deletedhistory'):
            if self.mw_version < '1.34':
                raise UserRightsError(err + 'deleted revisions.')
            if 'comment' in prop or 'parsedcomment' in prop:
                raise UserRightsError(err + 'comments of deleted revisions.')
        if ('content' in prop and not (self.has_right('deletedtext')
                                       or self.has_right('undelete'))):
            raise UserRightsError(err + 'deleted content.')

    def deletedrevs(
        self,
        titles: str
        | pywikibot.Page
        | Iterable[str]
        | Iterable[pywikibot.Page]
        | None = None,
        start=None,
        end=None,
        reverse: bool = False,
        content: bool = False,
        total: int | None = None,
        **kwargs,
    ) -> Generator[dict[str, Any], None, None]:
        """Iterate deleted revisions.

        Each value returned by the iterator will be a dict containing the
        'title' and 'ns' keys for a particular Page and a 'revisions' key
        whose value is a list of revisions in the same format as
        recentchanges plus a 'content' element with key '*' if requested
        when 'content' parameter is set. For older wikis a 'token' key is
        also given with the content request.

        .. seealso:: :api:`Deletedrevisions`

        :param titles: The page titles to check for deleted revisions
        :keyword revids: Get revisions by their ID

        .. note:: either titles or revids must be set but not both

        :param start: Iterate revisions starting at this Timestamp
        :param end: Iterate revisions ending at this Timestamp
        :param reverse: Iterate oldest revisions first (default: newest)
        :param content: If True, retrieve the content of each revision
        :param total: number of revisions to retrieve
        :keyword user: List revisions by this user
        :keyword excludeuser: Exclude revisions by this user
        :keyword tag: Only list revision tagged with this tag
        :keyword prop: Which properties to get. Defaults are ids, user,
            comment, flags and timestamp
        """
        # set default properties
        prop = kwargs.pop('prop',
                          ['ids', 'user', 'comment', 'flags', 'timestamp'])
        if isinstance(prop, str):
            prop = prop.split('|')
        if content:
            prop.append('content')

        if start and end:
            self.assert_valid_iter_params('deletedrevs', start, end, reverse)

        self._check_view_deleted('deletedrevs', prop)

        revids = kwargs.pop('revids', None)
        if not bool(titles) ^ (revids is not None):
            raise Error('deletedrevs: either "titles" or "revids" parameter '
                        'must be given.')

        gen = self._generator(api.PropertyGenerator,
                              type_arg='deletedrevisions',
                              titles=titles, revids=revids, total=total)

        gen.request['drvstart'] = start
        gen.request['drvend'] = end
        gen.request['drvprop'] = prop
        if reverse:
            gen.request['drvdir'] = 'newer'

        # handle other parameters like user
        for k, v in kwargs.items():
            gen.request['drv' + k] = v

        for data in gen:
            with suppress(KeyError):
                data['revisions'] = data.pop('deletedrevisions')
                yield data

    def alldeletedrevisions(
        self,
        *,
        namespaces: NamespaceArgType = None,
        reverse: bool = False,
        content: bool = False,
        total: int | None = None,
        **kwargs,
    ) -> Generator[dict[str, Any], None, None]:
        """Yield all deleted revisions.

        .. seealso:: :api:`Alldeletedrevisions`
        .. warning:: *user* keyword argument must be given together with
           *start* or *end*.

        :param namespaces: Only iterate pages in these namespaces
        :param reverse: Iterate oldest revisions first (default: newest)
        :param content: If True, retrieve the content of each revision
        :param total: Number of revisions to retrieve

        :keyword str from: Start listing at this title
        :keyword str to: Stop listing at this title
        :keyword str prefix: Search for all page titles that begin with this
            value
        :keyword str excludeuser: Exclude revisions by this user
        :keyword str tag: Only list revisions tagged with this tag
        :keyword str user: List revisions by this user
        :keyword start: Iterate revisions starting at this Timestamp
        :keyword end: Iterate revisions ending at this Timestamp
        :keyword list[str] prop: Which properties to get. Defaults are
            ``ids``, ``timestamp``, ``flags``, ``user``, and ``comment``
            (if the bot has the right to view).
        """
        if 'start' in kwargs and 'end' in kwargs:
            self.assert_valid_iter_params('alldeletedrevisions',
                                          kwargs['start'],
                                          kwargs['end'],
                                          reverse)
        prop = kwargs.pop('prop', [])
        parameters = {'adr' + k: v for k, v in kwargs.items()}
        if not prop:
            prop = ['ids', 'timestamp', 'flags', 'user']
            if self.has_right('deletedhistory'):
                prop.append('comment')
        if content:
            prop.append('content')
        self._check_view_deleted('alldeletedrevisions', prop)
        parameters['adrprop'] = prop
        if reverse:
            parameters['adrdir'] = 'newer'
        yield from self._generator(api.ListGenerator,
                                   type_arg='alldeletedrevisions',
                                   namespaces=namespaces,
                                   total=total,
                                   parameters=parameters)

    def users(
        self,
        usernames: Iterable[str],
    ) -> Iterable[dict[str, Any]]:
        """Iterate info about a list of users by name or IP.

        .. seealso:: :api:`Users`

        :param usernames: a list of user names
        """
        usprop = ['blockinfo', 'gender', 'groups', 'editcount', 'registration',
                  'rights', 'emailable']
        return api.ListGenerator(
            'users', site=self,
            parameters={'ususers': usernames, 'usprop': usprop}
        )

    def randompages(
        self,
        total: int | None = None,
        namespaces: NamespaceArgType = None,
        redirects: bool | None = False,
        content: bool = False,
    ) -> Iterable[pywikibot.Page]:
        """Iterate a number of random pages.

        Pages are listed in a fixed sequence, only the starting point is
        random.

        .. seealso:: :api:`Random`
        .. versionchanged:: 9.0
           Raises ``TypeError`` instead of ``AssertionError`` if
           *redirects* is invalid.

        :param total: the maximum number of pages to iterate
        :param namespaces: only iterate pages in these namespaces.
        :param redirects: if True, include only redirect pages in
            results, False does not include redirects and None include
            both types (default: False).
        :param content: if True, load the current content of each
            iterated page (default False).
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        :raises TypeError: unsupported redirects parameter
        """
        mapping = {False: None, True: 'redirects', None: 'all'}
        if redirects not in mapping:
            raise TypeError(f"Invalid type for 'redirects' parameter: "
                            f'{type(redirects).__name__}({redirects})')
        redirects_ = mapping[redirects]
        params = {}
        if redirects_ is not None:
            params['grnfilterredir'] = redirects_
        return self._generator(api.PageGenerator, type_arg='random',
                               namespaces=namespaces, total=total,
                               g_content=content, **params)

    _patrol_errors = {
        'nosuchrcid': 'There is no change with rcid {rcid}',
        'nosuchrevid': 'There is no change with revid {revid}',
        'patroldisabled': 'Patrolling is disabled on {site} wiki',
        'noautopatrol': 'User {user} has no permission to patrol its own '
                        'changes, "autopatrol" is needed',
        'notpatrollable':
            "The revision {revid} can't be patrolled as it's too old."
    }

    @need_right('patrol')
    def patrol(
        self,
        rcid: int | str | Iterable[int] | Iterable[str] | None = None,
        revid: int | str | Iterable[int] | Iterable[str] | None = None,
        revision: pywikibot.page.Revision
        | Iterable[pywikibot.page.Revision]
        | None = None,
    ) -> Generator[dict[str, int | str], None, None]:
        """Return a generator of patrolled pages.

        .. seealso:: :api:`Patrol`

        Pages to be patrolled are identified by rcid, revid or revision.
        At least one of the parameters is mandatory.
        See https://www.mediawiki.org/wiki/API:Patrol.

        :param rcid: an int/string/iterable/iterator providing rcid of pages
            to be patrolled.
        :param revid: an int/string/iterable/iterator providing revid of pages
            to be patrolled.
        :param revision: an Revision/iterable/iterator providing Revision
            object of pages to be patrolled.
        """
        # If patrol is not enabled, attr will be set the first time a
        # request is done.
        if hasattr(self, '_patroldisabled') and self._patroldisabled:
            return

        if all(_ is None for _ in [rcid, revid, revision]):
            raise Error('No rcid, revid or revision provided.')

        if rcid is None:
            rcid_ = set()
        elif isinstance(rcid, (int, str)):
            rcid_ = {rcid}
        else:
            rcid_ = set(rcid)
        if revid is None:
            revid_ = set()
        elif isinstance(revid, (int, str)):
            revid_ = {revid}
        else:
            revid_ = set(revid)
        if revision is None:
            revision_ = set()
        elif isinstance(revision, pywikibot.page.Revision):
            revision_ = {revision}
        else:
            revision_ = set(revision)

        combined_revid = revid_ | {r.revid for r in revision_}  # type: ignore[attr-defined]  # noqa: E501

        gen = itertools.chain(
            zip_longest(rcid_, [], fillvalue='rcid'),
            zip_longest(combined_revid, [], fillvalue='revid'))

        token = self.tokens['patrol']

        for idvalue, idtype in gen:
            req = self._request(parameters={'action': 'patrol',
                                            'token': token,
                                            idtype: idvalue})

            try:
                result = req.submit()
            except APIError as err:
                # patrol is disabled, store in attr to avoid other requests
                if err.code == 'patroldisabled':
                    self._patroldisabled = True
                    return

                errdata = {
                    'site': self,
                    'user': self.user(),
                    idtype: idvalue,
                }
                if err.code in self._patrol_errors:
                    raise Error(self._patrol_errors[err.code]
                                .format_map(errdata))
                pywikibot.debug(
                    f"protect: Unexpected error code '{err.code}' received."
                )
                raise

            yield result['patrol']

    def newpages(
        self,
        user=None,
        returndict: bool = False,
        start=None,
        end=None,
        reverse: bool = False,
        bot: bool = False,
        redirect: bool = False,
        excludeuser=None,
        patrolled=None,
        namespaces: NamespaceArgType = None,
        total: int | None = None,
    ) -> (
        Generator[tuple[pywikibot.Page, dict[str, Any]], None, None]
        | Generator[tuple[pywikibot.Page, str, int, str, str, str], None, None]
    ):
        """Yield new articles (as Page objects) from recent changes.

        Starts with the newest article and fetches the number of
        articles specified in the first argument.

        The objects yielded are dependent on parameter returndict. When
        true, it yields a tuple composed of a Page object and a dict of
        attributes. When false, it yields a tuple composed of the Page
        object, timestamp (str), length (int), an empty string, username
        or IP address (str), comment (str).

        :param namespaces: only iterate pages in these namespaces
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        # TODO: update docstring

        # N.B. API still provides no way to access Special:Newpages content
        # directly, so we get new pages indirectly through 'recentchanges'

        gen = self.recentchanges(
            start=start, end=end, reverse=reverse,
            namespaces=namespaces, changetype='new', user=user,
            excludeuser=excludeuser, bot=bot,
            redirect=redirect, patrolled=patrolled,
            total=total
        )
        for pageitem in gen:
            newpage = pywikibot.Page(self, pageitem['title'])
            if returndict:
                yield (newpage, pageitem)
            else:
                yield (newpage, pageitem['timestamp'], pageitem['newlen'],
                       '', pageitem['user'], pageitem.get('comment', ''))

    def querypage(
        self,
        special_page: str,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Iterate Page objects retrieved from Special:{special_page}.

        Generic function for all special pages supported by the site MW
        API.

        .. seealso:: :api:`Querypage`
        .. versionchanged:: 9.0
           Raises ``ValueError`` instead of ``AssertionError`` if
           *special_page* is invalid.

        :param special_page: Special page to query
        :param total: number of pages to return
        :raise ValueError: special_page is not supported in SpecialPages.
        """
        param = self._paraminfo.parameter('query+querypage', 'page')
        if special_page not in param['type']:
            raise ValueError(f"{special_page} not in {param['type']}")

        return self._generator(api.PageGenerator,
                               type_arg='querypage', gqppage=special_page,
                               total=total)

    def longpages(
        self,
        total: int | None = None,
    ) -> Generator[tuple[pywikibot.Page, int], None, None]:
        """Yield Pages and lengths from Special:Longpages.

        Yields a tuple of Page object, length(int).

        :param total: number of pages to return
        """
        lpgen = self._generator(api.ListGenerator,
                                type_arg='querypage', qppage='Longpages',
                                total=total)
        for pageitem in lpgen:
            yield (pywikibot.Page(self, pageitem['title']),
                   int(pageitem['value']))

    def shortpages(
        self,
        total: int | None = None,
    ) -> Generator[tuple[pywikibot.Page, int], None, None]:
        """Yield Pages and lengths from Special:Shortpages.

        Yields a tuple of Page object, length(int).

        :param total: number of pages to return
        """
        spgen = self._generator(api.ListGenerator,
                                type_arg='querypage', qppage='Shortpages',
                                total=total)
        for pageitem in spgen:
            yield (pywikibot.Page(self, pageitem['title']),
                   int(pageitem['value']))

    def deadendpages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Page objects retrieved from Special:Deadendpages.

        :param total: number of pages to return
        """
        return self.querypage('Deadendpages', total)

    def ancientpages(
        self,
        total: int | None = None,
    ) -> Generator[tuple[pywikibot.Page, pywikibot.Timestamp], None, None]:
        """Yield Pages, datestamps from Special:Ancientpages.

        :param total: number of pages to return
        """
        apgen = self._generator(api.ListGenerator,
                                type_arg='querypage', qppage='Ancientpages',
                                total=total)
        for pageitem in apgen:
            yield (pywikibot.Page(self, pageitem['title']),
                   pywikibot.Timestamp.fromISOformat(pageitem['timestamp']))

    def lonelypages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages retrieved from Special:Lonelypages.

        :param total: number of pages to return
        """
        return self.querypage('Lonelypages', total)

    def unwatchedpages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Unwatchedpages (requires Admin privileges).

        :param total: number of pages to return
        """
        return self.querypage('Unwatchedpages', total)

    def wantedpages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Wantedpages.

        :param total: number of pages to return
        """
        return self.querypage('Wantedpages', total)

    def wantedfiles(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Wantedfiles.

        :param total: number of pages to return
        """
        return self.querypage('Wantedfiles', total)

    def wantedtemplates(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Wantedtemplates.

        :param total: number of pages to return
        """
        return self.querypage('Wantedtemplates', total)

    def wantedcategories(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Wantedcategories.

        :param total: number of pages to return
        """
        return self.querypage('Wantedcategories', total)

    def uncategorizedcategories(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Categories from Special:Uncategorizedcategories.

        :param total: number of pages to return
        """
        return self.querypage('Uncategorizedcategories', total)

    def uncategorizedimages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield FilePages from Special:Uncategorizedimages.

        :param total: number of pages to return
        """
        return self.querypage('Uncategorizedimages', total)

    # synonym
    uncategorizedfiles = uncategorizedimages

    def uncategorizedpages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Uncategorizedpages.

        :param total: number of pages to return
        """
        return self.querypage('Uncategorizedpages', total)

    def uncategorizedtemplates(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages from Special:Uncategorizedtemplates.

        :param total: number of pages to return
        """
        return self.querypage('Uncategorizedtemplates', total)

    def unusedcategories(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Category objects from Special:Unusedcategories.

        :param total: number of pages to return
        """
        return self.querypage('Unusedcategories', total)

    def unusedfiles(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield FilePage objects from Special:Unusedimages.

        :param total: number of pages to return
        """
        return self.querypage('Unusedimages', total)

    def withoutinterwiki(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages without language links from Special:Withoutinterwiki.

        :param total: number of pages to return
        """
        return self.querypage('Withoutinterwiki', total)

    def broken_redirects(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages with broken redirects from Special:BrokenRedirects.

        :param total: number of pages to return
        """
        return self.querypage('BrokenRedirects', total)

    def double_redirects(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield Pages with double redirects from Special:DoubleRedirects.

        :param total: number of pages to return
        """
        return self.querypage('DoubleRedirects', total)

    def redirectpages(
        self,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Yield redirect pages from Special:ListRedirects.

        :param total: number of pages to return
        """
        return self.querypage('Listredirects', total)

    @deprecate_arg('type', 'protect_type')  # since 9.0
    def protectedpages(
        self,
        namespace: NamespaceArgType = 0,
        protect_type: str = 'edit',
        level: str | bool = False,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Return protected pages depending on protection level and type.

        For protection types which aren't 'create' it uses
        :py:obj:`APISite.allpages`, while it uses for 'create' the
        'query+protectedtitles' module.

        .. versionchanged:: 9.0
           *type* parameter was renamed to *protect_type*.

        .. seealso:: :api:`Protectedtitles`

        :param namespace: The searched namespace.
        :param protect_type: The protection type to search for
            (default 'edit').
        :param level: The protection level (like 'autoconfirmed'). If
            False it shows all protection levels.
        :return: The pages which are protected.
        """
        namespaces = self.namespaces.resolve(namespace)
        # always assert, so we are be sure that protect_type could be 'create'
        assert 'create' in self.restrictions['types'], \
            "'create' should be a valid protection type."
        if protect_type == 'create':
            return self._generator(
                api.PageGenerator, type_arg='protectedtitles',
                namespaces=namespaces, gptlevel=level, total=total)
        return self.allpages(namespace=namespaces[0], protect_level=level,
                             protect_type=protect_type, total=total)

    def pages_with_property(
        self,
        propname: str,
        *,
        total: int | None = None,
    ) -> Iterable[pywikibot.Page]:
        """Iterate Page objects from Special:PagesWithProp.

        .. seealso:: :api:`Pageswithprop`

        :param propname: must be a valid property.
        :param total: number of pages to return
        :return: return a generator of Page objects
        :rtype: iterator
        """
        if propname not in self.get_property_names():
            raise NotImplementedError(
                f'"{propname}" is not a valid page property')
        return self._generator(api.PageGenerator, type_arg='pageswithprop',
                               gpwppropname=propname, total=total)

    def watched_pages(
        self,
        force: bool = False,
        total: int | None = None,
        *,
        with_talkpage: bool = True,
    ) -> Iterable[pywikibot.Page]:
        """Return watchlist.

        .. note:: ``watched_pages`` is a restartable generator. See
           :class:`tools.collections.GeneratorWrapper` for its usage.
        .. seealso:: :api:`Watchlistraw`
        .. versionadded:: 8.1
           the *with_talkpage* parameter.

        :param force: Reload watchlist
        :param total: if not None, limit the generator to yielding this many
            items in total
        :param with_talkpage: if false, ignore talk pages and special
            pages
        :return: generator of pages in watchlist
        """
        def ignore_talkpages(page: pywikibot.page.BasePage) -> bool:
            """Ignore talk pages and special pages."""
            ns = page.namespace()
            return ns >= 0 and not page.namespace() % 2

        expiry = None if force else pywikibot.config.API_config_expiry
        gen = api.PageGenerator(site=self, generator='watchlistraw',
                                expiry=expiry)
        gen.set_maximum_items(total)

        if not with_talkpage:
            gen._namespaces = True
            gen._check_result_namespace = ignore_talkpages

        return gen



"""File containing all standard fixes."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pathlib import Path

from pywikibot import config


parameter_help = """
                  Currently available predefined fixes:

                  * HTML        - Convert HTML tags to wiki syntax, and
                                  fix XHTML.
                  * isbn        - Fix badly formatted ISBNs.
                  * syntax      - Try to fix bad wiki markup. Do not run
                                  this in automatic mode, as the bot may
                                  make mistakes.
                  * syntax-safe - Like syntax, but less risky; can be run
                                  in automatic mode.
                  * case-de     - Fix upper/lower case errors in German.
                  * grammar-de  - Fix grammar and typography in German.
                  * vonbis      - Replace hyphens or dashes with "bis"
                                  in German.
                  * music       - Links to disambiguation pages in German.
                  * datum       - Specific date formats in German.
                  * correct-ar  - Typo corrections for Arabic Wikipedia
                                  and other Arabic wikis.
                  * yu-tld      - Fix links to .yu domains, which are disabled.
                                  See:
                                  https://lists.wikimedia.org/pipermail/wikibots-l/2009-February/000290.html
                  * fckeditor   - Convert FCKeditor HTML tags to wiki syntax.
"""

__doc__ += parameter_help

fixes = {
    # These replacements will convert HTML to wiki syntax where possible, and
    # make remaining tags XHTML compliant.
    'HTML': {
        'regex': True,
        'msg': 'pywikibot-fixes-html',
        'replacements': [
            # Everything case-insensitive (?i)
            # Keep in mind that MediaWiki automatically converts <br> to <br />
            # when rendering pages, so you might comment the next two lines out
            # to save some time/edits.
            (r'(?i)<br *>',                      r'<br />'),
            # linebreak with attributes
            (r'(?i)<br ([^>/]+?)>',            r'<br \1 />'),
            (r'(?i)<b>(.*?)</b>',              r"'''\1'''"),
            (r'(?i)<strong>(.*?)</strong>',    r"'''\1'''"),
            (r'(?i)<i>(.*?)</i>',              r"''\1''"),
            (r'(?i)<em>(.*?)</em>',            r"''\1''"),
            # horizontal line without attributes in a single line
            (r'(?i)([\r\n])<hr[ /]*>([\r\n])', r'\1----\2'),
            # horizontal line without attributes with more text in same line
            #   (r'(?i) +<hr[ /]*> +',             r'\r\n----\r\n'),
            # horizontal line with attributes; can't be done with wiki syntax
            # so we only make it XHTML compliant
            (r'(?i)<hr ([^>/]+?)>',            r'<hr \1 />'),
            # a header where only spaces are in the same line
            (r'(?i)([\r\n]) *<h1> *([^<]+?) *</h1> *([\r\n])', r'\1= \2 =\3'),
            (r'(?i)([\r\n]) *<h2> *([^<]+?) *</h2> *([\r\n])',
             r'\1== \2 ==\3'),
            (r'(?i)([\r\n]) *<h3> *([^<]+?) *</h3> *([\r\n])',
             r'\1=== \2 ===\3'),
            (r'(?i)([\r\n]) *<h4> *([^<]+?) *</h4> *([\r\n])',
             r'\1==== \2 ====\3'),
            (r'(?i)([\r\n]) *<h5> *([^<]+?) *</h5> *([\r\n])',
             r'\1===== \2 =====\3'),
            (r'(?i)([\r\n]) *<h6> *([^<]+?) *</h6> *([\r\n])',
             r'\1====== \2 ======\3'),
            # TODO: maybe we can make the bot replace <p> tags with \r\n's.
        ],
        'exceptions': {
            'inside-tags': [
                'nowiki',
                'comment',
                'math',
                'pre',
                'syntaxhighlight',
            ],
        }
    },

    # Grammar fixes for German language
    # Do NOT run this automatically!
    'grammar-de': {
        'regex': True,
        'msg': {
            'de': 'Bot: korrigiere Grammatik',
        },
        'replacements': [
            # Vorsicht bei Substantiven, z. B. 3-Jähriger!
            (r'(\d+)(minütig|stündig|tägig|wöchig|jährig|minütlich|stündlich'
             r'|täglich|wöchentlich|jährlich|fach|mal|malig|köpfig|teilig'
             r'|gliedrig|geteilt|elementig|dimensional|bändig|eckig|farbig'
             r'|stimmig)', r'\1-\2'),
            # zusammengesetztes Wort, Bindestrich wird durchgeschleift
            (r'(?<!\w)(\d+|\d+[.,]\d+)(\$|€|DM|£|¥|mg|g|kg|ml|cl|l|t|ms|min'
             r'|µm|mm|cm|dm|m|km|ha|°C|kB|MB|GB|TB|W|kW|MW|GW|PS|Nm|eV|kcal'
             r'|mA|mV|kV|Ω|Hz|kHz|MHz|GHz|mol|Pa|Bq|Sv|mSv)([²³]?-[\w\[])',
             r'\1-\2\3'),
            # Größenangabe ohne Leerzeichen vor Einheit
            # weggelassen wegen vieler falsch Positiver: s, A, V, C, S, J, %
            (r'(?<!\w)(\d+|\d+[.,]\d+)(\$|€|DM|£|¥|mg|g|kg|ml|cl|l|t|ms|min'
             r'|µm|mm|cm|dm|m|km|ha|°C|kB|MB|GB|TB|W|kW|MW|GW|PS|Nm|eV|kcal'
             r'|mA|mV|kV|Ω|Hz|kHz|MHz|GHz|mol|Pa|Bq|Sv|mSv)(?=\W|²|³|$)',
             r'\1 \2'),
            # Temperaturangabe mit falsch gesetztem Leerzeichen
            (r'(?<!\w)(\d+|\d+[.,]\d+)° C(?=\W|²|³|$)', r'\1 °C'),
            # Kein Leerzeichen nach Komma
            (r'([a-zäöüß](\]\])?,)((\[\[)?[a-zäöüA-ZÄÖÜ])', r'\1 \3'),
            # Leerzeichen und Komma vertauscht
            (r'([a-zäöüß](\]\])?) ,((\[\[)?[a-zäöüA-ZÄÖÜ])', r'\1, \3'),
            # Plenks (Leerzeichen vor Komma/Punkt/Ausrufezeichen/Fragezeichen)
            # Achtung bei Französisch:
            # https://de.wikipedia.org/wiki/Plenk#Franz.C3.B6sische_Sprache
            # Leerzeichen vor Doppelpunkt/Semikolon kann korrekt sein,
            # z.B. nach Quellenangaben
            (r'([a-zäöüß](\]\])?) ([,.!?]) ((\[\[)?[a-zäöüA-ZÄÖÜ])',
             r'\1\3 \4'),
        ],
        'exceptions': {
            'inside-tags': [
                'nowiki',
                'comment',
                'gallery',          # because of filenames
                'hyperlink',        # e.g. commas in URLs
                'math',
                'pre',              # because of code examples
                'startspace',       # because of code examples
                'syntaxhighlight',  # because of code examples
                'timeline',
            ],
            'text-contains': [
                r'sic!',
                r'20min.ch',     # Schweizer News-Seite
            ],
            'inside': [
                r'<code>.*</code>',  # because of code examples
                r'{{[Zz]itat\|.*?}}',
                r'{{§\|.*?}}',   # Gesetzesparagraph
                r'§?\d+[a-z]',  # Gesetzesparagraph
                r'Ju 52/1m',  # Flugzeugbezeichnung
                r'Ju 52/3m',  # Flugzeugbezeichnung
                r'AH-1W',     # Hubschrauberbezeichnung
                r'ZPG-3W',    # Luftschiffbezeichnung
                r'8mm',       # Filmtitel
                r'802.11g',   # WLAN-Standard
                r'DOS/4GW',   # Software
                r'ntfs-3g',   # Dateisystem-Treiber
                r'/\w(,\w)*/',      # Laut-Aufzählung in der Linguistik
                # Variablen in der Mathematik
                # (unklar, ob Leerzeichen hier Pflicht sind)
                r'[xyz](,[xyz])+',
                # Definitionslisten, dort gibt es oft absichtlich Leerzeichen
                # vor Doppelpunkten
                r'(?m)^;(.*?)$',
                r'\d+h( |&nbsp;)\d+m',
                # Schreibweise für Zeiten, vor allem in Film-Infoboxen.
                # Nicht korrekt, aber dafür schön kurz.
                r'(?i)\[\[(Bild|Image|Media):.+?\|',  # Dateinamen auslassen
                r'{{bgc\|.*?}}',                      # Hintergrundfarbe
                r'<sup>\d+m</sup>',                   # bei chemischen Formeln
                r'\([A-Z][A-Za-z]*(,[A-Z][A-Za-z]*'
                r'(<sup>.*?</sup>|<sub>.*?</sub>|))+\)'
                # chemische Formel, z. B. AuPb(Pb,Sb,Bi)Te.
                # Hier sollen keine Leerzeichen hinter die Kommata.
            ],
            'title': [
                r'Arsen',  # chemische Formel
            ],
        }
    },

    # Do NOT run this automatically!
    # Recommendation: First run syntax-safe automatically, afterwards
    # run syntax manually, carefully checking that you're not breaking
    # anything.
    'syntax': {
        'regex': True,
        'msg': 'pywikibot-fixes-syntax',
        'replacements': [
            # external link in double brackets
            (r'\[\[(?P<url>https?://[^\]]+?)\]\]',   r'[\g<url>]'),
            # external link starting with double bracket
            (r'\[\[(?P<url>https?://.+?)\]',   r'[\g<url>]'),
            # external link with forgotten closing bracket
            #   (r'\[(?P<url>https?://[^\]\s]+)\r\n',  r'[\g<url>]\r\n'),
            # external link ending with double bracket.
            # do not change weblinks that contain wiki links inside
            # inside the description
            (r'\[(?P<url>https?://[^\[\]]+?)\]\](?!\])',   r'[\g<url>]'),
            # external link and description separated by a dash.
            # ATTENTION: while this is a mistake in most cases, there are some
            # valid URLs that contain dashes!
            (r'\[(?P<url>https?://[^\|\]\s]+?) *\| *(?P<label>[^\|\]]+?)\]',
             r'[\g<url> \g<label>]'),
            # wiki link closed by single bracket.
            # ATTENTION: There are some false positives, for example
            # Brainfuck code examples or MS-DOS parameter instructions.
            # There are also sometimes better ways to fix it than
            # just putting an additional ] after the link.
            (r'\[\[([^\[\]]+?)\](?!\])',  r'[[\1]]'),
            # wiki link opened by single bracket.
            # ATTENTION: same as above.
            (r'(?<!\[)\[([^\[\]]+?)\]\](?!\])',  r'[[\1]]'),
            # template closed by single bracket
            # ATTENTION: There are some false positives, especially in
            # mathematical context or program code.
            (r'{{([^{}]+?)}(?!})',       r'{{\1}}'),
        ],
        'exceptions': {
            'inside-tags': [
                'nowiki',
                'comment',
                'math',
                'pre',
                'startspace',       # because of code examples
                'syntaxhighlight',  # because of code examples
            ],
            'text-contains': [
                r'http://.*?object=tx\|',                # regular dash in URL
                r'http://.*?allmusic\.com',              # regular dash in URL
                r'http://.*?allmovie\.com',              # regular dash in URL
                r'http://physics.nist.gov/',             # regular dash in URL
                r'http://www.forum-seniorenarbeit.de/',  # regular dash in URL
                r'http://kuenstlerdatenbank.ifa.de/',    # regular dash in URL
                r'&object=med',                          # regular dash in URL
                r'\[CDATA\['                             # lots of brackets
            ],
        }
    },

    # The same as syntax, but restricted to replacements that should
    # be safe to run automatically.
    'syntax-safe': {
        'regex': True,
        'msg': 'pywikibot-fixes-syntax',
        'replacements': [
            # external link in double brackets
            (r'\[\[(?P<url>https?://[^\]]+?)\]\]',   r'[\g<url>]'),
            # external link starting with double bracket
            (r'\[\[(?P<url>https?://.+?)\]',   r'[\g<url>]'),
            # external link with forgotten closing bracket
            #   (r'\[(?P<url>https?://[^\]\s]+)\r\n',   r'[\g<url>]\r\n'),
            # external link and description separated by a dash, with
            # whitespace in front of the dash, so that it is clear that
            # the dash is not a legitimate part of the URL.
            (r'\[(?P<url>https?://[^\|\] \r\n]+?) +\| *(?P<label>[^\|\]]+?)\]',
             r'[\g<url> \g<label>]'),
            # dash in external link, where the correct end of the URL can
            # be detected from the file extension. It is very unlikely that
            # this will cause mistakes.
            (r'\[(?P<url>https?://[^\|\] ]+?'
             r'(\.pdf|\.html|\.htm|\.php|\.asp|\.aspx|\.jsp)) *\|'
             r' *(?P<label>[^\|\]]+?)\]', r'[\g<url> \g<label>]'),
        ],
        'exceptions': {
            'inside-tags': [
                'nowiki',
                'comment',
                'math',
                'pre',
                'startspace',       # because of code examples
                'syntaxhighlight',  # because of code examples
            ],
        }
    },

    'case-de': {  # German upper / lower case issues
        'regex': True,
        'msg': {
            'de': 'Bot: Korrigiere Groß-/Kleinschreibung',
        },
        'replacements': [
            (r'\batlantische(r|n|) Ozean', r'Atlantische\1 Ozean'),
            (r'\bdeutsche(r|n|) Bundestag\b', r'Deutsche\1 Bundestag'),
            # Aufpassen, z. B. 'deutsche Bundestagswahl'
            (r'\bdeutschen Bundestags\b', r'Deutschen Bundestags'),
            (r'\bdeutsche(r|n|) Reich\b', r'Deutsche\1 Reich'),
            # Aufpassen, z. B. 'deutsche Reichsgrenzen'
            (r'\bdeutschen Reichs\b', r'Deutschen Reichs'),
            (r'\bdritte(n|) Welt(?!krieg)', r'Dritte\1 Welt'),
            (r'\bdreißigjährige(r|n|) Krieg', r'Dreißigjährige\1 Krieg'),
            (r'\beuropäische(n|) Gemeinschaft', r'Europäische\1 Gemeinschaft'),
            (r'\beuropäische(n|) Kommission', r'Europäische\1 Kommission'),
            (r'\beuropäische(n|) Parlament', r'Europäische\1 Parlament'),
            (r'\beuropäische(n|) Union', r'Europäische\1 Union'),
            (r'\berste(r|n|) Weltkrieg', r'Erste\1 Weltkrieg'),
            (r'\bkalte(r|n|) Krieg', r'Kalte\1 Krieg'),
            (r'\bpazifische(r|n|) Ozean', r'Pazifische\1 Ozean'),
            (r'Tag der deutschen Einheit', r'Tag der Deutschen Einheit'),
            (r'\bzweite(r|n|) Weltkrieg', r'Zweite\1 Weltkrieg'),
        ],
        'exceptions': {
            'inside-tags': [
                'nowiki',
                'comment',
                'math',
                'pre',
            ],
            'text-contains': [
                r'sic!',
            ],
        }
    },

    'vonbis': {
        'regex': True,
        'msg': {
            'de': 'Bot: Ersetze Binde-/Gedankenstrich durch "bis"',
        },
        'replacements': [
            # Bindestrich, Gedankenstrich, Geviertstrich
            (r'(von \d{3,4}) *(-|&ndash;|–|&mdash;|—) *(\d{3,4})',
             r'\1 bis \3'),
        ],
    },

    # some disambiguation stuff for de:
    # python pwb.py replace -fix:music -subcat:Album
    'music': {
        'regex': False,
        'msg': {
            'de': 'Bot: korrigiere Links auf Begriffsklärungen',
        },
        'replacements': [
            ('[[CD]]', '[[Audio-CD|CD]]'),
            ('[[LP]]', '[[Langspielplatte|LP]]'),
            ('[[EP]]', '[[Extended Play|EP]]'),
            ('[[MC]]', '[[Musikkassette|MC]]'),
            ('[[Single]]', '[[Single (Musik)|Single]]'),
        ],
        'exceptions': {
            'inside-tags': [
                'hyperlink',
            ]
        }
    },

    # format of dates of birth and death, for de:
    # python pwb.py replace -fix:datum -ref:Vorlage:Personendaten
    'datum': {
        'regex': True,
        'msg': {
            'de': 'Bot: Korrigiere Datumsformat',
        },
        'replacements': [
            (r'\[\[(\d+\. (?:Januar|Februar|März|April|Mai|Juni|Juli|August|'
             r'September|Oktober|November|Dezember)) (\d{1,4})\]\]',
             r'[[\1]] [[\2]]'),
            # Keine führende Null beim Datum
            # (erst einmal nur bei fehlenden Leerzeichen)
            (r'0(\d+)\.(Januar|Februar|März|April|Mai|Juni|Juli|August|'
             r'September|Oktober|November|Dezember)', r'\1. \2'),
            # Kein Leerzeichen zwischen Tag und Monat
            (r'(\d+)\.(Januar|Februar|März|April|Mai|Juni|Juli|August|'
             r'September|Oktober|November|Dezember)', r'\1. \2'),
            # Kein Punkt vorm Jahr
            (r'(\d+)\. (Januar|Februar|März|April|Mai|Juni|Juli|August|'
             r'September|Oktober|November|Dezember)\.(\d{1,4})', r'\1. \2 \3'),
        ],
        'exceptions': {
            'inside': [
                r'\[\[20. Juli 1944\]\]',  # Hitler-Attentat
                r'\[\[17. Juni 1953\]\]',  # Ost-Berliner Volksaufstand
                r'\[\[1. April 2000\]\]',  # Film
                r'\[\[11. September 2001\]\]',  # Anschläge in den USA
                r'\[\[7. Juli 2005\]\]',   # Terroranschläge in Spanien
            ],
        }
    },

    'isbn': {
        'generator': [
            r'-search:insource:/nowiki\>ISBN:? *(?:&nbsp;|&\#160;)? *[0-9]/',
            '-namespace:0'],
        'regex': True,
        'msg': 'pywikibot-fixes-isbn',
        'replacements': [
            # Remove colon between the word ISBN and the number
            (r'ISBN: (\d+)', r'ISBN \1'),
            # superfluous word "number"
            (r'ISBN(?: [Nn]umber| [Nn]o\.?|-Nummer|-Nr\.):? (\d+)',
             r'ISBN \1'),
            # Space, minus, dot, hyphen, en dash, em dash, etc. instead of
            # hyphen-minus as separator,
            # or spaces between digits and separators.
            # Note that these regular expressions also match valid ISBNs, but
            # these won't be changed.
            # These two regexes don't verify that the ISBN is of a valid format
            # but just change separators into normal hyphens. The isbn script
            # does checks and similar but does only match ISBNs with digits and
            # hyphens (and optionally a X/x at the end).
            (r'ISBN (978|979) *[\- −.‐-―] *(\d+) *[\- −.‐-―] *(\d+) '
             r'*[\- −.‐-―] *(\d+) *[\- −.‐-―] *(\d)(?!\d)',
             r'ISBN \1-\2-\3-\4-\5'),  # ISBN-13

            (r'ISBN (\d+) *[\- −.‐-―] *(\d+) *[\- −.‐-―] *(\d+) *'
             r'[\- −.‐-―] *(\d|X|x)(?!\d)',
             r'ISBN \1-\2-\3-\4'),  # ISBN-10
            # missing space before ISBN-10 or before ISBN-13,
            # or multiple spaces or non-breaking space.
            (r'ISBN(?: *|&nbsp;)((\d(-?)){12}\d|(\d(-?)){9}[\dXx])',
             r'ISBN \1'),
            # remove <nowiki /> tags
            (r'<nowiki>ISBN:? *(?:&nbsp;|&#160;)? *([0-9\-xX]+)</nowiki>',
             r'ISBN \1'),
        ],
        'exceptions': {
            'inside-tags': [
                'comment',
                'hyperlink',
            ],
            'inside': [
                r'ISBN (97[89]-?)(\d-?){9}\d',  # matches valid ISBN-13s
                r'ISBN (\d-?){9}[\dXx]',  # matches valid ISBN-10s
            ],
        }
    },

    # Typo corrections for Arabic Wikipedia and any Arabic wiki.
    # python pwb.py replace -fix:correct-ar -start:! -always

    'correct-ar': {
        'regex': False,
        'msg': {
            'ar': 'بوت: تدقيق إملائي',
        },
        'replacements': [
            (' إمرأة ', ' امرأة '),
            (' الى ', ' إلى '),
            (' إسم ', ' اسم '),
            (' الأن ', ' الآن '),
            (' اول ', ' أول '),
            (' الة ', ' آلة '),
            (' فى ', ' في '),
            (' اثقل ', ' أثقل '),
            (' إبن ', ' ابن '),
            (' إبنة ', ' ابنة '),
            (' إقتصاد ', ' اقتصاد '),
            (' إجتماع ', ' اجتماع '),
            (' انجيل ', ' إنجيل '),
            (' اجماع ', ' إجماع '),
            (' امريكا ', ' أمريكا '),
            (' اوروبا ', ' أوروبا '),
            (' انجلترا ', ' إنجلترا '),
            (' اكتوبر ', ' أكتوبر '),
            (' اسرائيل ', ' إسرائيل '),
            (' المانيا ', ' ألمانيا '),
            (' ايطاليا ', ' إيطاليا '),
            (' ايران ', ' إيران '),
            (' إستخراج ', ' استخراج '),
            (' إستعمال ', ' استعمال '),
            (' إستبدال ', ' استبدال '),
            (' إشتراك ', ' اشتراك '),
            (' إستعادة ', ' استعادة '),
            (' إستقلال ', ' استقلال '),
            (' إنتقال ', ' انتقال '),
            (' إتحاد ', ' اتحاد '),
            (' الإتحاد ', ' الاتحاد '),
            (' املاء ', ' إملاء '),
            (' إستخدام ', ' استخدام '),
            (' أحدى ', ' إحدى '),
            (' لاكن ', ' لكن '),
            (' الاردن ', ' الأردن '),
            (' إثنان ', ' اثنان '),
            (' شيئ ', ' شيء '),
            (' إحتياط ', ' احتياط '),
            (' إقتباس ', ' اقتباس '),
            (' الامارات ', ' الإمارات '),
            (' اكثر ', ' أكثر '),
            (' افضل ', ' أفضل '),
            (' اكبر ', ' أكبر '),
            (' اشهر ', ' أشهر '),
            (' ادارة ', ' إدارة '),
            (' ابناء ', ' أبناء '),
            (' الانصار ', '  الأنصار '),
            (' اشارة ', ' إشارة '),
            (' إقرأ ', ' اقرأ '),
            (' إمتياز ', ' امتياز '),
            (' ارق ', ' أرق '),
            (' أرثوذوكس ', ' أرثوذكس '),
            (' الأرثوذوكس ', ' الأرثوذكس '),
            (' أرثوذوكسية ', ' أرثوذكسية '),
            (' الأرثوذوكسية ', ' الأرثوذكسية '),
            (' الأرثوذوكسي ', ' الأرثوذكسي '),
            (' ارثوذوكس ', ' أرثوذكس '),
            (' ارثوذوكسي ', ' أرثوذكسي '),
            (' ارثوذوكسية ', ' أرثوذكسية '),
            (' الارثوذوكسية ', ' الأرثوذكسية '),
            (' اللة ', ' الله '),
            (' إختبار ', ' اختبار '),
            (' ارسال ', ' إرسال '),
            (' إتصالات ', ' اتصالات '),
            (' اسامة ', ' أسامة '),
            (' ابراهيم ', ' إبراهيم '),
            (' اسماعيل ', ' إسماعيل '),
            (' ايوب ', ' أيوب '),
            (' ايمن ', ' أيمن '),
            (' ابو ', ' أبو '),
            (' ابا ', ' أبا '),
            (' اخو ', ' أخو '),
            (' اخا ', ' أخا '),
            (' اخي ', ' أخي '),
            (' احد ', ' أحد '),
            (' اربعاء ', ' أربعاء '),
            (' اهم ', ' أهم '),
            (' اوزبكستان ', ' أوزبكستان '),
            (' اذربيجان ', ' أذربيجان '),
            (' افغانستان ', ' أفغانستان '),
            (' امجد ', ' أمجد '),
            (' اوسط ', ' أوسط '),
            (' اشقر ', ' أشقر '),
            (' انور ', ' أنور '),
            (' اصعب ', ' أصعب '),
            (' اسهل ', ' أسهل '),
            (' اجمل ', ' أجمل '),
            (' اقبح ', ' أقبح '),
            (' اطول ', ' أطول '),
            (' اقصر ', ' أقصر '),
            (' اسمن ', ' أسمن '),
            (' اذكى ', ' أذكى '),
            (' اماني ', ' أماني '),
            (' احلام ', ' أحلام '),
            (' اسماء ', ' أسماء '),
            (' ابطأ ', ' أبطأ '),
            (' اوربا ', ' أوروبا '),
            (' أوربا ', ' أوروبا '),
            (' امريكي ', ' أمريكي '),
            (' امريكية ', ' أمريكية '),
            (' امريكيان ', ' أمريكيان '),
            (' امريكيتان ', ' أمريكيتان '),
            (' امريكيون ', ' أمريكيون '),
            (' امريكيات ', ' أمريكيات '),
            (' الامريكي ', ' الأمريكي '),
            (' الامريكية ', ' الأمريكية '),
            (' الامريكيان ', ' الأمريكيان '),
            (' الامريكيتان ', ' الأمريكيتان '),
            (' الامريكيون ', ' الأمريكيون '),
            (' الامريكيات ', ' الأمريكيات '),
            (' اوروبي ', ' أوروبي '),
            (' اوروبية ', ' أوروبية '),
            (' اوروبيان ', ' أوروبيان '),
            (' اوروبيتان ', ' أوروبيتان '),
            (' اوروبيون ', ' أوروبيون '),
            (' اوروبيات ', ' أوروبيات '),
            (' الاوروبي ', ' الأوروبي '),
            (' الاوروبية ', ' الأوروبية '),
            (' الاوروبيان ', ' الأوروبيان '),
            (' الاوروبيتان ', ' الأوروبيتان '),
            (' الاوروبيون ', ' الأوروبيون '),
            (' الاوروبيات ', ' الأوروبيات '),
            (' اسرائيلي ', ' إسرائيلي '),
            (' اسرائيلية ', ' إسرائيلية '),
            (' اسرائيليان ', ' إسرائيليان '),
            (' اسرائيليتان ', ' إسرائيليتان '),
        ],
        'exceptions': {
            'inside-tags': [
                'file',         # because of filenames
                'gallery',      # because of filenames
                'header',       # section headers
                'hyperlink',    # URLs
                'interwiki',
                'ref',
            ],
        }
    },
    # TODO: Support dynamic replacement from Special pages to the localized one
    'specialpages': {
        'regex': False,
        'msg': {
            'ar': 'روبوت: إصلاح حالة حروف الصفحات الخاصة',
            'arz': 'روبوت: تصليح حاله الحروف بتاعه الصفحات الخصوصيه',
            'en': 'Robot: Fixing special page capitalisation',
            'fa': 'ربات: تصحیح بزرگی و کوچکی حروف صفحه‌های ویژه',
        },
        'replacements': [
            ('Special:Allpages',        'Special:AllPages'),
            ('Special:Blockip',         'Special:BlockIP'),
            ('Special:Blankpage',       'Special:BlankPage'),
            ('Special:Filepath',        'Special:FilePath'),
            ('Special:Globalusers',     'Special:GlobalUsers'),
            ('Special:Imagelist',       'Special:ImageList'),
            ('Special:Ipblocklist',     'Special:IPBlockList'),
            ('Special:Listgrouprights', 'Special:ListGroupRights'),
            ('Special:Listusers',       'Special:ListUsers'),
            ('Special:Newimages',       'Special:NewImages'),
            ('Special:Prefixindex',     'Special:PrefixIndex'),
            ('Special:Protectedpages',  'Special:ProtectedPages'),
            ('Special:Recentchanges',   'Special:RecentChanges'),
            ('Special:Specialpages',    'Special:SpecialPages'),
            ('Special:Unlockdb',        'Special:UnlockDB'),
            ('Special:Userlogin',       'Special:UserLogin'),
            ('Special:Userlogout',      'Special:UserLogout'),
            ('Special:Whatlinkshere',   'Special:WhatLinksHere'),
        ],
    },
    # yu top-level domain will soon be disabled, see
    # https://lists.wikimedia.org/pipermail/wikibots-l/2009-February/000290.html
    # The following are domains that are often-used.
    'yu-tld': {
        'regex': False,
        'nocase': True,
        'msg': {
            'ar':  'روبوت: إصلاح الوصلات إلى نطاقات .yu',
            'arz':  'روبوت: تصليح اللينكات للدومين .yu',
            'de':  'Bot: Ersetze Links auf .yu-Domains',
            'en':  'Robot: Replacing links to .yu domains',
            'fa':  'ربات: جایگزینی پیوندها به دامنه‌ها با پسوند yu',
            'fr':  ('Robot: Correction des liens pointant vers le domaine '
                    '.yu, qui expire en 2009'),
            'ksh': 'Bot: de ahle .yu-Domains loufe us, dröm ußjetuusch',
            'sr': 'Бот: Исправљање линкова ка .yu домену',
        },
        'replacements': [
            ('www.budva.cg.yu',             'www.budva.rs'),
            ('spc.org.yu',                  'spc.rs'),
            ('www.oks.org.yu',              'www.oks.org.rs'),
            ('www.kikinda.org.yu',          'www.kikinda.rs'),
            ('www.ds.org.yu',               'www.ds.org.rs'),
            ('www.nbs.yu',                  'www.nbs.rs'),
            ('www.serbia.sr.gov.yu',        'www.srbija.gov.rs'),
            ('eunet.yu',                    'eunet.rs'),
            ('www.zastava-arms.co.yu',      'www.zastava-arms.co.rs'),
            ('www.airportnis.co.yu',        'www.airportnis.rs'),
            ('www.belex.co.yu',             'www.belex.rs'),
            ('beograd.org.yu',              'beograd.rs'),
            ('www.vlada.cg.yu',             'www.vlada.me'),
            ('webrzs.statserb.sr.gov.yu',   'webrzs.stat.gov.rs'),
            ('www.statserb.sr.gov.yu',      'webrzs.stat.gov.rs'),
            ('www.rastko.org.yu',           'www.rastko.org.rs'),
            ('www.reprezentacija.co.yu',    'www.reprezentacija.rs'),
            ('www.blic.co.yu',              'www.blic.co.rs'),
            ('www.beograd.org.yu',          'www.beograd.org.rs'),
            ('arhiva.glas-javnosti.co.yu',  'arhiva.glas-javnosti.rs'),
            ('www.srpsko-nasledje.co.yu',   'www.srpsko-nasledje.co.rs'),
            ('www.dnevnik.co.yu',           'www.dnevnik.rs'),
            ('www.srbija.sr.gov.yu',        'www.srbija.gov.rs'),
            ('www.kurir-info.co.yu/Arhiva', 'arhiva.kurir-info.rs/Arhiva'),
            ('www.kurir-info.co.yu/arhiva', 'arhiva.kurir-info.rs/arhiva'),
            ('www.kurir-info.co.yu',        'www.kurir-info.rs'),
            ('arhiva.kurir-info.co.yu',     'arhiva.kurir-info.rs'),
            ('www.prvaliga.co.yu',          'www.prvaliga.rs'),
            ('www.mitropolija.cg.yu',       'www.mitropolija.me'),
            ('www.spc.yu/sr',               'www.spc.rs/sr'),
            ('www.sk.co.yu',                'www.sk.co.rs'),
            ('www.ekoforum.org.yu',         'www.ekoforum.org'),
            ('www.svevlad.org.yu',          'www.svevlad.org.rs'),
            ('www.posta.co.yu',             'www.posta.rs'),
            ('www.glas-javnosti.co.yu',     'www.glas-javnosti.rs'),
            ('www.fscg.cg.yu',              'www.fscg.co.me'),
            ('ww1.rts.co.yu/euro',          'ww1.rts.co.rs/euro'),
            ('www.rtv.co.yu',               'www.rtv.rs'),
            ('www.politika.co.yu',          'www.politika.rs'),
            ('www.mfa.gov.yu',              'www.mfa.gov.rs'),
            ('www.drzavnauprava.sr.gov.yu', 'www.drzavnauprava.gov.rs'),
        ],
    },
    # These replacements will convert HTML tag from FCK-editor to wiki syntax.
    #
    'fckeditor': {
        'regex': True,
        'msg': 'pywikibot-fixes-fckeditor',
        'replacements': [
            # replace <br> with a new line
            (r'(?i)<br>', r'\n'),
            # replace &nbsp; with a space
            (r'(?i)&nbsp;', r' '),
        ],
    },
}


def _load_file(filename: str) -> bool:
    """Load the fixes from the given filename.

    Returns True if the file existed and was loaded, False otherwise.

    :meta public:
    """
    path = Path(filename)
    if path.exists():
        # Read file as binary, so that compile can detect encoding from header
        with path.open('rb') as f:
            code = compile(f.read(), filename, 'exec')
            exec(code, globals())  # intentionally in globals
        return True

    return False


# Load the user fixes file.
filename = config.datafilepath('user-fixes.py')
user_fixes_loaded = _load_file(filename)



"""Pywikibot metadata file.

.. versionadded:: 4.0
"""
#
# (C) Pywikibot team, 2020-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from time import strftime


__version__ = '10.7.4'
__url__ = 'https://www.mediawiki.org/wiki/Manual:Pywikibot'
__copyright__ = f'2003-{strftime("%Y")}, Pywikibot team'



"""Objects representing a MediaWiki page.

Various Wikibase pages are defined in ``page._wikibase.py``,
various pages for Proofread Extensions are defined in
``pywikibot.proofreadpage``.

.. note:: `Link` objects represent a wiki-page's title, while
   :class:`pywikibot.Page` objects (defined here) represent the page
   itself, including its contents.
"""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot
from pywikibot import textlib
from pywikibot.exceptions import (
    Error,
    InterwikiRedirectPageError,
    IsNotRedirectPageError,
    IsRedirectPageError,
    NoPageError,
    UnknownExtensionError,
)
from pywikibot.page._basepage import BasePage
from pywikibot.page._toolforge import WikiBlameMixin
from pywikibot.site import Namespace
from pywikibot.tools import cached, deprecated_args


__all__ = ['Page']


class Page(BasePage, WikiBlameMixin):

    """Page: A MediaWiki page."""

    def __init__(self, source, title: str = '', ns=0) -> None:
        """Instantiate a Page object."""
        if isinstance(source, pywikibot.site.BaseSite) and not title:
            raise ValueError('Title must be specified and not empty '
                             'if source is a Site.')
        super().__init__(source, title, ns)

    @property
    @cached
    def raw_extracted_templates(self):
        """Extract templates and parameters.

        This method is using
        :func:`textlib.extract_templates_and_params`.
        Disabled parts and whitespace are stripped, except for
        whitespace in anonymous positional arguments.

        :rtype: list of (str, OrderedDict)
        """
        return textlib.extract_templates_and_params(self.text, True, True)

    def templatesWithParams(  # noqa: N802
        self,
    ) -> list[tuple[pywikibot.Page, list[str]]]:
        """Return templates used on this Page.

        The templates are extracted by :meth:`raw_extracted_templates`,
        with positional arguments placed first in order, and each named
        argument appearing as 'name=value'.

        All parameter keys and values for each template are stripped of
        whitespace.

        :return: a list of tuples with one tuple for each template
            invocation in the page, with the template Page as the first
            entry and a list of parameters as the second entry.
        """
        # WARNING: may not return all templates used in particularly
        # intricate cases such as template substitution
        titles = {t.title()
                  for t in self.templates(namespaces=Namespace.TEMPLATE)}
        templates = self.raw_extracted_templates
        # backwards-compatibility: convert the dict returned as the second
        # element into a list in the format used by old scripts
        result = []
        for template in templates:
            try:
                link = pywikibot.Link(template[0], self.site,
                                      default_namespace=Namespace.TEMPLATE)
                if link.canonical_title() not in titles:
                    continue
            except Error:
                # this is a parser function or magic word, not template name
                # the template name might also contain invalid parts
                continue
            args = template[1]
            intkeys = {}
            named = {}
            positional = []
            for key in sorted(args):
                try:
                    intkeys[int(key)] = args[key]
                except ValueError:
                    named[key] = args[key]

            for i in range(1, len(intkeys) + 1):
                # only those args with consecutive integer keys can be
                # treated as positional; an integer could also be used
                # (out of order) as the key for a named argument
                # example: {{tmp|one|two|5=five|three}}
                if i in intkeys:
                    positional.append(intkeys[i])
                    continue

                for k, v in intkeys.items():
                    if k < 1 or k >= i:
                        named[str(k)] = v
                break

            positional += [f'{key}={value}' for key, value in named.items()]
            result.append((pywikibot.Page(link, self.site), positional))
        return result

    @deprecated_args(botflag='bot')  # since 9.3.0
    def set_redirect_target(
        self,
        target_page: pywikibot.Page | str,
        create: bool = False,
        force: bool = False,
        keep_section: bool = False,
        save: bool = True,
        **kwargs
    ) -> None:
        """Change the page's text to point to the redirect page.

        .. versionchanged:: 9.3
           *botflag* keyword parameter was renamed to *bot*.

        :param target_page: target of the redirect, this argument is
            required.
        :param create: if true, it creates the redirect even if the page
            doesn't exist.
        :param force: if true, it set the redirect target even the page
            doesn't exist or it's not redirect.
        :param keep_section: if the old redirect links to a section
            and the new one doesn't it uses the old redirect's section.
        :param save: if true, it saves the page immediately.
        :param kwargs: Arguments which are used for saving the page
            directly afterwards, like *summary* for edit summary.
        """
        if isinstance(target_page, str):
            target_page = pywikibot.Page(self.site, target_page)
        elif self.site != target_page.site:
            raise InterwikiRedirectPageError(self, target_page)

        if not self.exists() and not (create or force):
            raise NoPageError(self)

        if self.exists() and not self.isRedirectPage() and not force:
            raise IsNotRedirectPageError(self)

        old_text = self.text
        result = self.site.redirect_regex.search(old_text)
        if result:
            oldlink = result[1]
            if (keep_section and '#' in oldlink
                    and target_page.section() is None):
                sectionlink = oldlink[oldlink.index('#'):]
                target_page = pywikibot.Page(
                    self.site,
                    target_page.title() + sectionlink
                )
            prefix = self.text[:result.start()]
            suffix = self.text[result.end():]
        else:
            prefix = suffix = ''

        target_link = target_page.title(as_link=True, textlink=True,
                                        allow_interwiki=False)
        target_link = f'#{self.site.redirect()} {target_link}'
        self.text = prefix + target_link + suffix
        if save:
            self.save(**kwargs)

    def get_best_claim(self, prop: str) -> pywikibot.Claim | None:
        """Return the first best Claim for this page.

        Return the first 'preferred' ranked Claim specified by Wikibase
        property or the first 'normal' one otherwise.

        .. versionadded:: 3.0

        .. seealso:: :meth:`pywikibot.ItemPage.get_best_claim`

        :param prop: Wikibase property ID, must be of the form ``P``
            followed by one or more digits (e.g. ``P31``).
        :return: Claim object given by Wikibase property number
            for this page object.

        :raises UnknownExtensionError: site has no Wikibase extension
        """
        def get_item_page(page):
            if not page.site.has_data_repository:
                raise UnknownExtensionError(
                    f'Wikibase is not implemented for {page.site}.')
            try:
                item_p = page.data_item()
                item_p.get()
                return item_p
            except NoPageError:
                return None
            except IsRedirectPageError:
                return get_item_page(item_p.getRedirectTarget())

        item_page = get_item_page(page=self)
        if item_page:
            return item_page.get_best_claim(prop)
        return None



"""Objects with site methods independent of the communication interface."""
#
# (C) Pywikibot team, 2008-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import functools
import inspect
import re
import threading
from warnings import warn

import pywikibot
from pywikibot.backports import Pattern
from pywikibot.exceptions import (
    Error,
    FamilyMaintenanceWarning,
    NoPageError,
    PageInUseError,
    UnknownSiteError,
)
from pywikibot.site._namespace import Namespace, NamespacesDict
from pywikibot.throttle import Throttle
from pywikibot.tools import (
    ComparableMixin,
    cached,
    deprecated,
    first_upper,
    normalize_username,
)


class BaseSite(ComparableMixin):

    """Site methods that are independent of the communication interface."""

    def __init__(self, code: str, fam=None, user=None) -> None:
        """Initializer.

        :param code: the site's language code
        :type code: str
        :param fam: wiki family name (optional)
        :type fam: str or pywikibot.family.Family
        :param user: bot user name (optional)
        :type user: str
        """
        if code.lower() != code:
            # Note the Site function in __init__ also emits a UserWarning
            # for this condition, showing the callers file and line no.
            pywikibot.log(f'BaseSite: code "{code}" converted to lowercase')
            code = code.lower()
        if not all(x in pywikibot.family.CODE_CHARACTERS for x in code):
            pywikibot.log(
                f'BaseSite: code "{code}" contains invalid characters')
        self.__code = code
        if isinstance(fam, str) or fam is None:
            self.__family = pywikibot.family.Family.load(fam)
        else:
            self.__family = fam

        self.obsolete = False
        # if we got an outdated language code, use the new one instead.
        if self.__code in self.__family.obsolete:
            if self.__family.obsolete[self.__code] is not None:
                self.__code = self.__family.obsolete[self.__code]
                # Note the Site function in __init__ emits a UserWarning
                # for this condition, showing the callers file and line no.
                pywikibot.log(
                    f'Site {self} instantiated using aliases code of {code}')
            else:
                # no such language anymore
                self.obsolete = True
                pywikibot.log(f'Site {self} instantiated and marked "obsolete"'
                              ' to prevent access')
        elif self.__code not in self.codes:
            if self.__family.name in self.__family.langs \
               and len(self.__family.langs) == 1:
                self.__code = self.__family.name
                if self.__family == pywikibot.config.family \
                   and code == pywikibot.config.mylang:
                    pywikibot.config.mylang = self.__code
                    warn('Global configuration variable "mylang" changed to '
                         f'"{self.__code}" while instantiating site {self}',
                         UserWarning, stacklevel=2)
            else:
                error_msg = (f"Language '{self.__code}' does not exist in "
                             f'family {self.__family.name}')
                raise UnknownSiteError(error_msg)

        self._username = normalize_username(user)

        # following are for use with lock_page and unlock_page methods
        self._pagemutex = threading.Condition()
        self._locked_pages: set[str] = set()

    @property
    @deprecated(since='8.5.0')
    def use_hard_category_redirects(self) -> bool:
        """Hard redirects are used for this site.

        Originally create as property for future use for a proposal to
        replace category redirect templates with hard redirects. This
        was never implemented and is not used inside the framework.

        .. deprecated:: 8.5
        """
        return False

    @property
    @cached
    def throttle(self):
        """Return this Site's throttle.

        Initialize a new one if needed.
        """
        return Throttle(self)

    @property
    def family(self):
        """The Family object for this Site's wiki family."""
        return self.__family

    @property
    def code(self):
        """The identifying code for this Site equal to the wiki prefix.

        By convention, this is usually an ISO language code, but it does
        not have to be.
        """
        return self.__code

    @property
    def lang(self):
        """The ISO language code for this Site.

        Presumed to be equal to the site code, but this can be
        overridden.
        """
        return self.__code

    @property
    @cached
    def doc_subpage(self) -> tuple:
        """Return the documentation subpage for this Site."""
        try:
            doc, codes = self.family.doc_subpages.get('_default', ((), []))
            if self.code not in codes:
                try:
                    doc = self.family.doc_subpages[self.code]
                # Language not defined in doc_subpages in x_family.py file
                # It will use default for the family.
                # should it just raise an Exception and fail?
                # this will help to check the dictionary ...
                except KeyError:
                    warn(f'Site {self} has no language defined in '
                         f'doc_subpages dict in {self.family.name}_family.py '
                         'file', FamilyMaintenanceWarning, 2)
        # doc_subpages not defined in x_family.py file
        except AttributeError:
            doc = ()  # default
            warn(f'Site {self} has no doc_subpages dict in '
                 f'{self.family.name}_family.py file',
                 FamilyMaintenanceWarning, 2)

        return doc

    def _cmpkey(self):
        """Perform equality and inequality tests on Site objects."""
        return (self.family.name, self.code)

    def __getstate__(self):
        """Remove Lock based classes before pickling."""
        new = self.__dict__.copy()
        del new['_pagemutex']
        if '_throttle' in new:
            del new['_throttle']
        # site cache contains exception information, which can't be pickled
        if '_iw_sites' in new:
            del new['_iw_sites']
        return new

    def __setstate__(self, attrs) -> None:
        """Restore things removed in __getstate__."""
        self.__dict__.update(attrs)
        self._pagemutex = threading.Condition()

    def user(self) -> str | None:
        """Return the currently-logged in bot username, or None."""
        if self.logged_in():
            return self.username()
        return None

    def username(self) -> str | None:
        """Return the username used for the site."""
        return self._username

    def __getattr__(self, name: str):
        """Delegate undefined methods calls to the Family object.

        .. versionchanged:: 9.0
           Only delegate to public Family methods which have ``code`` as
           first parameter.
        """
        if not name.startswith('_'):
            obj = getattr(self.family, name, None)
            if inspect.ismethod(obj):
                params = inspect.signature(obj).parameters
                if params:
                    parameter = next(iter(params))
                    if parameter == 'code':
                        method = functools.partial(obj, self.code)
                        if hasattr(obj, '__doc__'):
                            method.__doc__ = obj.__doc__
                        return method

        raise AttributeError(f'{type(self).__name__} instance has no '
                             f'attribute {name!r}') from None

    def __str__(self) -> str:
        """Return string representing this Site's name and code."""
        return self.family.name + ':' + self.code

    @property
    def sitename(self):
        """String representing this Site's name and code."""
        return str(self)

    def __repr__(self) -> str:
        """Return internal representation."""
        return f'{type(self).__name__}({self.code!r}, {self.family.name!r})'

    def __hash__(self):
        """Return hash value of instance."""
        return hash(repr(self))

    @deprecated('codes', since='9.6')
    def languages(self) -> list[str]:
        """Return list of all valid site codes for this site's Family.

        .. deprecated:: 9.6
           Use :meth:`codes` instead.
        """
        return sorted(self.codes)

    @property
    def codes(self) -> set[str]:
        """Return set of all valid site codes for this site's Family.

        .. versionadded:: 9.6
        .. seealso:: :attr:`family.Family.codes`
        """
        return set(self.family.langs.keys())

    def validLanguageLinks(self):  # noqa: N802
        """Return list of language codes to be used in interwiki links."""
        return [lang for lang in sorted(self.codes)
                if self.namespaces.lookup_normalized_name(lang) is None]

    def _interwiki_urls(self, only_article_suffixes: bool = False):
        base_path = self.path()
        if not only_article_suffixes:
            yield base_path + '{}'
        yield base_path + '/{}'
        yield base_path + '?title={}'
        yield self.articlepath

    @staticmethod
    def _build_namespaces():
        """Create default namespaces."""
        return Namespace.builtin_namespaces()

    @property
    @cached
    def namespaces(self):
        """Return dict of valid namespaces on this wiki."""
        return NamespacesDict(self._build_namespaces())

    def ns_normalize(self, value: str):
        """Return canonical local form of namespace name.

        :param value: A namespace name
        """
        index = self.namespaces.lookup_name(value)
        return self.namespace(index)

    def redirect(self) -> str:
        """Return a default redirect tag for the site.

        .. versionchanged:: 8.4
           return a single generic redirect tag instead of a list of
           tags. For the list use :meth:`redirects` instead.
        """
        return self.redirects()[0]

    def redirects(self) -> list[str]:
        """Return list of generic redirect tags for the site.

        .. seealso:: :meth:`redirect` for the default redirect tag.
        .. versionadded:: 8.4
        """
        return ['REDIRECT']

    def pagenamecodes(self) -> list[str]:
        """Return list of localized PAGENAME tags for the site."""
        return ['PAGENAME']

    def pagename2codes(self) -> list[str]:
        """Return list of localized PAGENAMEE tags for the site."""
        return ['PAGENAMEE']

    def lock_page(self, page, block: bool = True) -> None:
        """Lock page for writing. Must be called before writing any page.

        We don't want different threads trying to write to the same page
        at the same time, even to different sections.

        :param page: the page to be locked
        :type page: pywikibot.Page
        :param block: if true, wait until the page is available to be
            locked; otherwise, raise an exception if page can't be
            locked
        """
        title = page.title(with_section=False)
        with self._pagemutex:
            while title in self._locked_pages:
                if not block:
                    raise PageInUseError(title)
                self._pagemutex.wait()
            self._locked_pages.add(title)

    def unlock_page(self, page) -> None:
        """Unlock page. Call as soon as a write operation has completed.

        :param page: the page to be locked
        :type page: pywikibot.Page
        """
        with self._pagemutex:
            self._locked_pages.discard(page.title(with_section=False))
            self._pagemutex.notify_all()

    def disambcategory(self):
        """Return Category in which disambig pages are listed."""
        if self.has_data_repository:
            repo = self.data_repository()
            repo_name = repo.family.name
            try:
                item = self.family.disambcatname[repo.code]
            except KeyError:
                raise Error(f'No {repo_name} qualifier found for'
                            ' disambiguation category name in '
                            f'{self.family.name}_family file')

            dp = pywikibot.ItemPage(repo, item)
            try:
                name = dp.getSitelink(self)
            except NoPageError:
                raise Error(f'No disambiguation category name found in {repo} '
                            f'for {self}')

        else:  # fallback for non WM sites
            try:
                name = (f'{Namespace.CATEGORY}:'
                        f'{self.family.disambcatname[self.code]}')
            except KeyError:
                raise Error(f'No disambiguation category name found in '
                            f'{self.family.name}_family for {self}')

        return pywikibot.Category(pywikibot.Link(name, self))

    def isInterwikiLink(self, text):  # noqa: N802
        """Return True if text is in the form of an interwiki link.

        If a link object constructed using "text" as the link text
        parses as belonging to a different site, this method returns
        True.
        """
        linkfam, linkcode = pywikibot.Link(text, self).parse_site()
        return linkfam != self.family.name or linkcode != self.code

    @property
    def redirect_regex(self) -> Pattern[str]:
        """Return a compiled regular expression matching on redirect pages.

        Group 1 in the regex match object will be the target title.

        A redirect starts with hash (#), followed by a keyword, then
        arbitrary stuff, then a wikilink. The wikilink may contain a
        label, although this is not useful.

        .. versionadded:: 8.4
           moved from class:`APISite<pywikibot.site._apisite.APISite>`
        """
        tags = '|'.join(self.redirects())
        return re.compile(fr'\s*#(?:{tags})\s*:?\s*\[\[(.+?)(?:\|.*?)?\]\]',
                          re.IGNORECASE | re.DOTALL)

    def sametitle(self, title1: str, title2: str) -> bool:
        """Return True if title1 and title2 identify the same wiki page.

        title1 and title2 may be unequal but still identify the same
        page, if they use different aliases for the same namespace.
        """
        def ns_split(title):
            """Separate the namespace from the name."""
            ns, delim, name = title.partition(':')
            if delim:
                ns = self.namespaces.lookup_name(ns)
            if not delim or not ns:
                return default_ns, title
            return ns, name

        # Replace alias characters like underscores with title
        # delimiters like spaces and multiple combinations of them with
        # only one delimiter
        sep = self.family.title_delimiter_and_aliases[0]
        pattern = re.compile(f'[{self.family.title_delimiter_and_aliases}]+')
        title1 = pattern.sub(sep, title1)
        title2 = pattern.sub(sep, title2)
        if title1 == title2:
            return True

        default_ns = self.namespaces[0]
        # determine whether titles contain namespace prefixes
        ns1_obj, name1 = ns_split(title1)
        ns2_obj, name2 = ns_split(title2)
        if ns1_obj != ns2_obj:
            # pages in different namespaces
            return False

        name1 = name1.strip()
        name2 = name2.strip()
        # If the namespace has a case definition it's overriding the site's
        # case definition
        if ns1_obj.case == 'first-letter':
            name1 = first_upper(name1)
            name2 = first_upper(name2)
        return name1 == name2

    # site-specific formatting preferences

    def category_on_one_line(self):
        # TODO: is this even needed? No family in the framework uses it.
        """Return True if this site wants all category links on one line."""
        return self.code in self.family.category_on_one_line

    def interwiki_putfirst(self):
        """Return list of language codes for ordering of interwiki links."""
        return self.family.interwiki_putfirst.get(self.code)

    def getSite(self, code):  # noqa: N802
        """Return Site object for language 'code' in this Family."""
        return pywikibot.Site(code=code, fam=self.family, user=self.user())



"""Family module for Translate Wiki."""
#
# (C) Pywikibot team, 2007-2023
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


# The Wikimedia i18n family
class Family(family.SingleSiteFamily):

    """Family class for Translate Wiki."""

    name = 'i18n'
    domain = 'translatewiki.net'



"""Family module for Wikitech."""
#
# (C) Pywikibot team, 2005-2022
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


# The Wikitech family
class Family(family.WikimediaOrgFamily):

    """Family class for Wikitech."""

    name = 'wikitech'
    code = 'en'

    def protocol(self, code) -> str:
        """Return the protocol for this family."""
        return 'https'



#!/usr/bin/env python3
"""Tests for the tests package."""
#
# (C) Pywikibot team, 2014-2025
#
# Distributed under the terms of the MIT license.
from __future__ import annotations

import unittest
from contextlib import suppress

from tests import utils
from tests.aspects import DefaultSiteTestCase, TestCase, require_version


class HttpServerProblemTestCase(TestCase):

    """Test HTTP status 502 causes this test class to be skipped."""

    sites = {
        '502': {
            'hostname': 'http://httpbin.org/status/502',
        }
    }

    def test_502(self) -> None:  # pragma: no cover
        """Test that framework is skipping this test due to HTTP status 502."""
        self.fail("The test framework should skip this test but it hasn't.")


class TestLengthAssertion(TestCase):

    """Test length assertion methods.

    ``@unittest.expectedFailure`` is used to test the failure of a test;
    this is intentional. If the decorated test passes unexpectedly the
    test will fail.
    """

    net = False

    seq1 = ('foo', 'bar', 'baz')
    seq2 = 'foo'

    def test_assert_is_empty(self) -> None:
        """Test assertIsEmpty method."""
        self.assertIsEmpty([])
        self.assertIsEmpty('')

    @unittest.expectedFailure
    def test_assert_is_empty_fail(self) -> None:
        """Test assertIsEmpty method failing."""
        self.assertIsEmpty(self.seq1)
        self.assertIsEmpty(self.seq2)  # pragma: no cover

    def test_assert_is_not_empty(self) -> None:
        """Test assertIsNotEmpty method."""
        self.assertIsNotEmpty(self.seq1)
        self.assertIsNotEmpty(self.seq2)

    @unittest.expectedFailure
    def test_assert_is_not_empty_fail(self) -> None:
        """Test that assertIsNotEmpty method may fail."""
        self.assertIsNotEmpty([])
        self.assertIsNotEmpty('')  # pragma: no cover

    def test_assert_length(self) -> None:
        """Test assertLength method."""
        self.assertLength([], 0)
        self.assertLength(self.seq1, 3)
        self.assertLength(self.seq1, self.seq2)

    @unittest.expectedFailure
    def test_assert_length_fail(self) -> None:
        """Test that assertLength method is failing."""
        self.assertLength([], 1)
        # no cover: start
        self.assertLength(self.seq1, 0)
        self.assertLength(None, self.seq)
        # no cover: stop


class TestRequireVersionDry(DefaultSiteTestCase):

    """Test require_version decorator."""

    dry = True

    @require_version('')
    def method(self) -> None:
        """Test method for decorator."""

    def test_require_version(self) -> None:
        """Test require_version for DrySite."""
        with self.assertRaisesRegex(
            TypeError,
                f'{type(self).__name__}.site must be a BaseSite not DrySite'):
            self.method()


class TestRequireVersion(DefaultSiteTestCase):

    """Test require_version decorator."""

    @require_version('')
    def method_with_params(self, key) -> None:
        """Test method for decorated methods with unsupported arguments."""

    def method_failing(self) -> None:
        """Test method for decorator with invalid parameter."""
        self.assertTrue(False, 'should never happen')  # pragma: no cover

    @require_version('>=1.31')
    def method_succeed(self) -> None:
        """Test that decorator passes."""
        self.assertTrue(False, 'intentional fail for method_succeed test')

    @require_version('<1.31')
    def method_fail(self) -> None:
        """Test that decorator skips."""
        self.assertTrue(False, 'intentional fail for test')  # pragma: no cover

    def test_unsupported_methods(self) -> None:
        """Test require_version with unsupported methods."""
        with self.assertRaisesRegex(
                TypeError, "Test method 'method_with_params' has parameters"):
            self.method_with_params('42')
        with self.assertRaisesRegex(
                TypeError, "Test method 'method_with_params' has parameters"):
            self.method_with_params(key='42')
        with self.assertRaisesRegex(ValueError,
                                    'There is no valid operator given '):
            self.method_with_params()

    def test_version_needed(self) -> None:
        """Test for invalid decorator parameters."""
        with self.assertRaisesRegex(ValueError,
                                    'There is no valid operator given '):
            require_version('foo')(self.method_failing)(self)
        with self.assertRaisesRegex(ValueError,
                                    'first operand foo should not be set'):
            require_version('foo>bar')(self.method_failing)(self)
        with self.assertRaisesRegex(ValueError, 'Invalid version number'):
            require_version('>bar')(self.method_failing)(self)
        with self.assertRaisesRegex(unittest.SkipTest,
                                    r'MediaWiki < v1\.31 required'):
            require_version('<1.31')(self.method_failing)(self)
        with self.assertRaisesRegex(
                unittest.SkipTest,
                r'MediaWiki < v1\.31 required to run this test'):
            require_version('<1.31',
                            'run this test')(self.method_failing)(self)

    def test_decorator(self) -> None:
        """Test that decorator passes or skips."""
        with self.assertRaisesRegex(
            AssertionError,
                'intentional fail for method_succeed test'):
            self.method_succeed()
        with self.assertRaisesRegex(unittest.SkipTest,
                                    r'MediaWiki < v1\.31 required'):
            self.method_fail()


class UtilsTests(TestCase):

    """Tests for tests.utils."""

    net = False
    pattern = 'Hello World'

    def test_fixed_generator(self) -> None:
        """Test utils.fixed_generator."""
        gen = utils.fixed_generator(self.pattern)
        self.assertEqual(list(gen(1, 'foo', bar='baz')), list(self.pattern))

    def test_entered_loop(self) -> None:
        """Test utils.entered_loop."""
        self.assertTrue(utils.entered_loop(self.pattern))
        self.assertFalse(utils.entered_loop(''))

    @utils.expected_failure_if(True)
    def test_expected_failure_true(self) -> None:
        """Test expected_failure_if decorator if condition is True."""
        self.assertTrue(False)

    @utils.expected_failure_if(False)
    def test_expected_failure_false(self) -> None:
        """Test expected_failure_if decorator if condition is False."""
        self.assertTrue(True)


if __name__ == '__main__':
    with suppress(SystemExit):
        unittest.main()
